#
# PySNMP MIB module IPMROUTE-STD-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file:///home/mvold/mibs/cisco/IPMROUTE-STD-MIB.my
# Produced by pysmi-0.3.4 at Thu Jul 18 12:37:03 2024
# On host agrajag platform Linux version 6.6.37 by user mvold
# Using Python version 3.11.9 (main, Apr  2 2024, 08:25:04) [GCC 13.2.0]
#
ObjectIdentifier, OctetString, Integer = mibBuilder.importSymbols("ASN1", "ObjectIdentifier", "OctetString", "Integer")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsIntersection, ConstraintsUnion, SingleValueConstraint, ValueSizeConstraint, ValueRangeConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsIntersection", "ConstraintsUnion", "SingleValueConstraint", "ValueSizeConstraint", "ValueRangeConstraint")
IANAipRouteProtocol, IANAipMRouteProtocol = mibBuilder.importSymbols("IANA-RTPROTO-MIB", "IANAipRouteProtocol", "IANAipMRouteProtocol")
InterfaceIndexOrZero, InterfaceIndex = mibBuilder.importSymbols("IF-MIB", "InterfaceIndexOrZero", "InterfaceIndex")
SnmpAdminString, = mibBuilder.importSymbols("SNMP-FRAMEWORK-MIB", "SnmpAdminString")
ModuleCompliance, ObjectGroup, NotificationGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ModuleCompliance", "ObjectGroup", "NotificationGroup")
NotificationType, Counter32, Integer32, MibScalar, MibTable, MibTableRow, MibTableColumn, IpAddress, Gauge32, iso, Unsigned32, ObjectIdentity, Counter64, TimeTicks, Bits, MibIdentifier, ModuleIdentity, mib_2 = mibBuilder.importSymbols("SNMPv2-SMI", "NotificationType", "Counter32", "Integer32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "IpAddress", "Gauge32", "iso", "Unsigned32", "ObjectIdentity", "Counter64", "TimeTicks", "Bits", "MibIdentifier", "ModuleIdentity", "mib-2")
TruthValue, TextualConvention, DisplayString, RowStatus = mibBuilder.importSymbols("SNMPv2-TC", "TruthValue", "TextualConvention", "DisplayString", "RowStatus")
ipMRouteStdMIB = ModuleIdentity((1, 3, 6, 1, 2, 1, 83))
ipMRouteStdMIB.setRevisions(('2000-09-22 00:00',))
if mibBuilder.loadTexts: ipMRouteStdMIB.setLastUpdated('200009220000Z')
if mibBuilder.loadTexts: ipMRouteStdMIB.setOrganization('IETF IDMR Working Group')
class LanguageTag(TextualConvention, OctetString):
    status = 'current'
    displayHint = '100a'
    subtypeSpec = OctetString.subtypeSpec + ValueSizeConstraint(1, 100)

ipMRouteMIBObjects = MibIdentifier((1, 3, 6, 1, 2, 1, 83, 1))
ipMRoute = MibIdentifier((1, 3, 6, 1, 2, 1, 83, 1, 1))
ipMRouteEnable = MibScalar((1, 3, 6, 1, 2, 1, 83, 1, 1, 1), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("enabled", 1), ("disabled", 2)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ipMRouteEnable.setStatus('current')
ipMRouteEntryCount = MibScalar((1, 3, 6, 1, 2, 1, 83, 1, 1, 7), Gauge32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteEntryCount.setStatus('current')
ipMRouteTable = MibTable((1, 3, 6, 1, 2, 1, 83, 1, 1, 2), )
if mibBuilder.loadTexts: ipMRouteTable.setStatus('current')
ipMRouteEntry = MibTableRow((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1), ).setIndexNames((0, "IPMROUTE-STD-MIB", "ipMRouteGroup"), (0, "IPMROUTE-STD-MIB", "ipMRouteSource"), (0, "IPMROUTE-STD-MIB", "ipMRouteSourceMask"))
if mibBuilder.loadTexts: ipMRouteEntry.setStatus('current')
ipMRouteGroup = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 1), IpAddress())
if mibBuilder.loadTexts: ipMRouteGroup.setStatus('current')
ipMRouteSource = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 2), IpAddress())
if mibBuilder.loadTexts: ipMRouteSource.setStatus('current')
ipMRouteSourceMask = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 3), IpAddress())
if mibBuilder.loadTexts: ipMRouteSourceMask.setStatus('current')
ipMRouteUpstreamNeighbor = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 4), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteUpstreamNeighbor.setStatus('current')
ipMRouteInIfIndex = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 5), InterfaceIndexOrZero()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteInIfIndex.setStatus('current')
ipMRouteUpTime = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 6), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteUpTime.setStatus('current')
ipMRouteExpiryTime = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 7), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteExpiryTime.setStatus('current')
ipMRoutePkts = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 8), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRoutePkts.setStatus('current')
ipMRouteDifferentInIfPackets = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 9), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteDifferentInIfPackets.setStatus('current')
ipMRouteOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 10), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteOctets.setStatus('current')
ipMRouteProtocol = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 11), IANAipMRouteProtocol()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteProtocol.setStatus('current')
ipMRouteRtProto = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 12), IANAipRouteProtocol()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteRtProto.setStatus('current')
ipMRouteRtAddress = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 13), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteRtAddress.setStatus('current')
ipMRouteRtMask = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 14), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteRtMask.setStatus('current')
ipMRouteRtType = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 15), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("unicast", 1), ("multicast", 2)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteRtType.setStatus('current')
ipMRouteHCOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 2, 1, 16), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteHCOctets.setStatus('current')
ipMRouteNextHopTable = MibTable((1, 3, 6, 1, 2, 1, 83, 1, 1, 3), )
if mibBuilder.loadTexts: ipMRouteNextHopTable.setStatus('current')
ipMRouteNextHopEntry = MibTableRow((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1), ).setIndexNames((0, "IPMROUTE-STD-MIB", "ipMRouteNextHopGroup"), (0, "IPMROUTE-STD-MIB", "ipMRouteNextHopSource"), (0, "IPMROUTE-STD-MIB", "ipMRouteNextHopSourceMask"), (0, "IPMROUTE-STD-MIB", "ipMRouteNextHopIfIndex"), (0, "IPMROUTE-STD-MIB", "ipMRouteNextHopAddress"))
if mibBuilder.loadTexts: ipMRouteNextHopEntry.setStatus('current')
ipMRouteNextHopGroup = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 1), IpAddress())
if mibBuilder.loadTexts: ipMRouteNextHopGroup.setStatus('current')
ipMRouteNextHopSource = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 2), IpAddress())
if mibBuilder.loadTexts: ipMRouteNextHopSource.setStatus('current')
ipMRouteNextHopSourceMask = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 3), IpAddress())
if mibBuilder.loadTexts: ipMRouteNextHopSourceMask.setStatus('current')
ipMRouteNextHopIfIndex = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 4), InterfaceIndex())
if mibBuilder.loadTexts: ipMRouteNextHopIfIndex.setStatus('current')
ipMRouteNextHopAddress = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 5), IpAddress())
if mibBuilder.loadTexts: ipMRouteNextHopAddress.setStatus('current')
ipMRouteNextHopState = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 6), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("pruned", 1), ("forwarding", 2)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteNextHopState.setStatus('current')
ipMRouteNextHopUpTime = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 7), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteNextHopUpTime.setStatus('current')
ipMRouteNextHopExpiryTime = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 8), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteNextHopExpiryTime.setStatus('current')
ipMRouteNextHopClosestMemberHops = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 9), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteNextHopClosestMemberHops.setStatus('current')
ipMRouteNextHopProtocol = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 10), IANAipMRouteProtocol()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteNextHopProtocol.setStatus('current')
ipMRouteNextHopPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 3, 1, 11), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteNextHopPkts.setStatus('current')
ipMRouteInterfaceTable = MibTable((1, 3, 6, 1, 2, 1, 83, 1, 1, 4), )
if mibBuilder.loadTexts: ipMRouteInterfaceTable.setStatus('current')
ipMRouteInterfaceEntry = MibTableRow((1, 3, 6, 1, 2, 1, 83, 1, 1, 4, 1), ).setIndexNames((0, "IPMROUTE-STD-MIB", "ipMRouteInterfaceIfIndex"))
if mibBuilder.loadTexts: ipMRouteInterfaceEntry.setStatus('current')
ipMRouteInterfaceIfIndex = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 4, 1, 1), InterfaceIndex())
if mibBuilder.loadTexts: ipMRouteInterfaceIfIndex.setStatus('current')
ipMRouteInterfaceTtl = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 4, 1, 2), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 255))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ipMRouteInterfaceTtl.setStatus('current')
ipMRouteInterfaceProtocol = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 4, 1, 3), IANAipMRouteProtocol()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteInterfaceProtocol.setStatus('current')
ipMRouteInterfaceRateLimit = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 4, 1, 4), Integer32()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ipMRouteInterfaceRateLimit.setStatus('current')
ipMRouteInterfaceInMcastOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 4, 1, 5), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteInterfaceInMcastOctets.setStatus('current')
ipMRouteInterfaceOutMcastOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 4, 1, 6), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteInterfaceOutMcastOctets.setStatus('current')
ipMRouteInterfaceHCInMcastOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 4, 1, 7), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteInterfaceHCInMcastOctets.setStatus('current')
ipMRouteInterfaceHCOutMcastOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 4, 1, 8), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ipMRouteInterfaceHCOutMcastOctets.setStatus('current')
ipMRouteBoundaryTable = MibTable((1, 3, 6, 1, 2, 1, 83, 1, 1, 5), )
if mibBuilder.loadTexts: ipMRouteBoundaryTable.setStatus('current')
ipMRouteBoundaryEntry = MibTableRow((1, 3, 6, 1, 2, 1, 83, 1, 1, 5, 1), ).setIndexNames((0, "IPMROUTE-STD-MIB", "ipMRouteBoundaryIfIndex"), (0, "IPMROUTE-STD-MIB", "ipMRouteBoundaryAddress"), (0, "IPMROUTE-STD-MIB", "ipMRouteBoundaryAddressMask"))
if mibBuilder.loadTexts: ipMRouteBoundaryEntry.setStatus('current')
ipMRouteBoundaryIfIndex = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 5, 1, 1), InterfaceIndex())
if mibBuilder.loadTexts: ipMRouteBoundaryIfIndex.setStatus('current')
ipMRouteBoundaryAddress = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 5, 1, 2), IpAddress())
if mibBuilder.loadTexts: ipMRouteBoundaryAddress.setStatus('current')
ipMRouteBoundaryAddressMask = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 5, 1, 3), IpAddress())
if mibBuilder.loadTexts: ipMRouteBoundaryAddressMask.setStatus('current')
ipMRouteBoundaryStatus = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 5, 1, 4), RowStatus()).setMaxAccess("readcreate")
if mibBuilder.loadTexts: ipMRouteBoundaryStatus.setStatus('current')
ipMRouteScopeNameTable = MibTable((1, 3, 6, 1, 2, 1, 83, 1, 1, 6), )
if mibBuilder.loadTexts: ipMRouteScopeNameTable.setStatus('current')
ipMRouteScopeNameEntry = MibTableRow((1, 3, 6, 1, 2, 1, 83, 1, 1, 6, 1), ).setIndexNames((0, "IPMROUTE-STD-MIB", "ipMRouteScopeNameAddress"), (0, "IPMROUTE-STD-MIB", "ipMRouteScopeNameAddressMask"), (1, "IPMROUTE-STD-MIB", "ipMRouteScopeNameLanguage"))
if mibBuilder.loadTexts: ipMRouteScopeNameEntry.setStatus('current')
ipMRouteScopeNameAddress = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 6, 1, 1), IpAddress())
if mibBuilder.loadTexts: ipMRouteScopeNameAddress.setStatus('current')
ipMRouteScopeNameAddressMask = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 6, 1, 2), IpAddress())
if mibBuilder.loadTexts: ipMRouteScopeNameAddressMask.setStatus('current')
ipMRouteScopeNameLanguage = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 6, 1, 3), LanguageTag())
if mibBuilder.loadTexts: ipMRouteScopeNameLanguage.setStatus('current')
ipMRouteScopeNameString = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 6, 1, 4), SnmpAdminString()).setMaxAccess("readcreate")
if mibBuilder.loadTexts: ipMRouteScopeNameString.setStatus('current')
ipMRouteScopeNameDefault = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 6, 1, 5), TruthValue().clone('false')).setMaxAccess("readcreate")
if mibBuilder.loadTexts: ipMRouteScopeNameDefault.setStatus('current')
ipMRouteScopeNameStatus = MibTableColumn((1, 3, 6, 1, 2, 1, 83, 1, 1, 6, 1, 6), RowStatus()).setMaxAccess("readcreate")
if mibBuilder.loadTexts: ipMRouteScopeNameStatus.setStatus('current')
ipMRouteMIBConformance = MibIdentifier((1, 3, 6, 1, 2, 1, 83, 2))
ipMRouteMIBCompliances = MibIdentifier((1, 3, 6, 1, 2, 1, 83, 2, 1))
ipMRouteMIBGroups = MibIdentifier((1, 3, 6, 1, 2, 1, 83, 2, 2))
ipMRouteMIBCompliance = ModuleCompliance((1, 3, 6, 1, 2, 1, 83, 2, 1, 1)).setObjects(("IPMROUTE-STD-MIB", "ipMRouteMIBBasicGroup"), ("IPMROUTE-STD-MIB", "ipMRouteMIBRouteGroup"), ("IPMROUTE-STD-MIB", "ipMRouteMIBBoundaryGroup"), ("IPMROUTE-STD-MIB", "ipMRouteMIBHCInterfaceGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ipMRouteMIBCompliance = ipMRouteMIBCompliance.setStatus('current')
ipMRouteMIBBasicGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 83, 2, 2, 1)).setObjects(("IPMROUTE-STD-MIB", "ipMRouteEnable"), ("IPMROUTE-STD-MIB", "ipMRouteEntryCount"), ("IPMROUTE-STD-MIB", "ipMRouteUpstreamNeighbor"), ("IPMROUTE-STD-MIB", "ipMRouteInIfIndex"), ("IPMROUTE-STD-MIB", "ipMRouteUpTime"), ("IPMROUTE-STD-MIB", "ipMRouteExpiryTime"), ("IPMROUTE-STD-MIB", "ipMRouteNextHopState"), ("IPMROUTE-STD-MIB", "ipMRouteNextHopUpTime"), ("IPMROUTE-STD-MIB", "ipMRouteNextHopExpiryTime"), ("IPMROUTE-STD-MIB", "ipMRouteNextHopProtocol"), ("IPMROUTE-STD-MIB", "ipMRouteNextHopPkts"), ("IPMROUTE-STD-MIB", "ipMRouteInterfaceTtl"), ("IPMROUTE-STD-MIB", "ipMRouteInterfaceProtocol"), ("IPMROUTE-STD-MIB", "ipMRouteInterfaceRateLimit"), ("IPMROUTE-STD-MIB", "ipMRouteInterfaceInMcastOctets"), ("IPMROUTE-STD-MIB", "ipMRouteInterfaceOutMcastOctets"), ("IPMROUTE-STD-MIB", "ipMRouteProtocol"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ipMRouteMIBBasicGroup = ipMRouteMIBBasicGroup.setStatus('current')
ipMRouteMIBHopCountGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 83, 2, 2, 2)).setObjects(("IPMROUTE-STD-MIB", "ipMRouteNextHopClosestMemberHops"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ipMRouteMIBHopCountGroup = ipMRouteMIBHopCountGroup.setStatus('current')
ipMRouteMIBBoundaryGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 83, 2, 2, 3)).setObjects(("IPMROUTE-STD-MIB", "ipMRouteBoundaryStatus"), ("IPMROUTE-STD-MIB", "ipMRouteScopeNameString"), ("IPMROUTE-STD-MIB", "ipMRouteScopeNameDefault"), ("IPMROUTE-STD-MIB", "ipMRouteScopeNameStatus"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ipMRouteMIBBoundaryGroup = ipMRouteMIBBoundaryGroup.setStatus('current')
ipMRouteMIBPktsOutGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 83, 2, 2, 4)).setObjects(("IPMROUTE-STD-MIB", "ipMRouteNextHopPkts"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ipMRouteMIBPktsOutGroup = ipMRouteMIBPktsOutGroup.setStatus('current')
ipMRouteMIBHCInterfaceGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 83, 2, 2, 5)).setObjects(("IPMROUTE-STD-MIB", "ipMRouteInterfaceHCInMcastOctets"), ("IPMROUTE-STD-MIB", "ipMRouteInterfaceHCOutMcastOctets"), ("IPMROUTE-STD-MIB", "ipMRouteHCOctets"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ipMRouteMIBHCInterfaceGroup = ipMRouteMIBHCInterfaceGroup.setStatus('current')
ipMRouteMIBRouteGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 83, 2, 2, 6)).setObjects(("IPMROUTE-STD-MIB", "ipMRouteRtProto"), ("IPMROUTE-STD-MIB", "ipMRouteRtAddress"), ("IPMROUTE-STD-MIB", "ipMRouteRtMask"), ("IPMROUTE-STD-MIB", "ipMRouteRtType"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ipMRouteMIBRouteGroup = ipMRouteMIBRouteGroup.setStatus('current')
ipMRouteMIBPktsGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 83, 2, 2, 7)).setObjects(("IPMROUTE-STD-MIB", "ipMRoutePkts"), ("IPMROUTE-STD-MIB", "ipMRouteDifferentInIfPackets"), ("IPMROUTE-STD-MIB", "ipMRouteOctets"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ipMRouteMIBPktsGroup = ipMRouteMIBPktsGroup.setStatus('current')
mibBuilder.exportSymbols("IPMROUTE-STD-MIB", ipMRouteNextHopUpTime=ipMRouteNextHopUpTime, ipMRouteSourceMask=ipMRouteSourceMask, ipMRouteScopeNameAddress=ipMRouteScopeNameAddress, ipMRouteNextHopSource=ipMRouteNextHopSource, ipMRouteNextHopPkts=ipMRouteNextHopPkts, ipMRouteNextHopClosestMemberHops=ipMRouteNextHopClosestMemberHops, ipMRouteInIfIndex=ipMRouteInIfIndex, ipMRouteUpTime=ipMRouteUpTime, ipMRouteMIBBoundaryGroup=ipMRouteMIBBoundaryGroup, ipMRouteStdMIB=ipMRouteStdMIB, ipMRouteUpstreamNeighbor=ipMRouteUpstreamNeighbor, ipMRouteInterfaceHCOutMcastOctets=ipMRouteInterfaceHCOutMcastOctets, ipMRouteRtAddress=ipMRouteRtAddress, ipMRouteBoundaryAddressMask=ipMRouteBoundaryAddressMask, ipMRouteScopeNameEntry=ipMRouteScopeNameEntry, ipMRouteInterfaceProtocol=ipMRouteInterfaceProtocol, ipMRouteInterfaceTable=ipMRouteInterfaceTable, ipMRouteProtocol=ipMRouteProtocol, ipMRouteEntryCount=ipMRouteEntryCount, ipMRouteScopeNameLanguage=ipMRouteScopeNameLanguage, ipMRoute=ipMRoute, ipMRouteMIBPktsOutGroup=ipMRouteMIBPktsOutGroup, ipMRouteNextHopEntry=ipMRouteNextHopEntry, ipMRouteMIBCompliances=ipMRouteMIBCompliances, ipMRouteHCOctets=ipMRouteHCOctets, ipMRouteBoundaryEntry=ipMRouteBoundaryEntry, ipMRouteMIBCompliance=ipMRouteMIBCompliance, ipMRouteInterfaceRateLimit=ipMRouteInterfaceRateLimit, ipMRouteNextHopAddress=ipMRouteNextHopAddress, ipMRouteExpiryTime=ipMRouteExpiryTime, ipMRouteInterfaceEntry=ipMRouteInterfaceEntry, ipMRouteDifferentInIfPackets=ipMRouteDifferentInIfPackets, ipMRouteInterfaceHCInMcastOctets=ipMRouteInterfaceHCInMcastOctets, LanguageTag=LanguageTag, ipMRouteMIBObjects=ipMRouteMIBObjects, ipMRouteBoundaryTable=ipMRouteBoundaryTable, ipMRouteMIBPktsGroup=ipMRouteMIBPktsGroup, ipMRoutePkts=ipMRoutePkts, ipMRouteMIBRouteGroup=ipMRouteMIBRouteGroup, ipMRouteMIBBasicGroup=ipMRouteMIBBasicGroup, ipMRouteBoundaryAddress=ipMRouteBoundaryAddress, ipMRouteRtProto=ipMRouteRtProto, ipMRouteOctets=ipMRouteOctets, ipMRouteMIBConformance=ipMRouteMIBConformance, ipMRouteBoundaryIfIndex=ipMRouteBoundaryIfIndex, ipMRouteNextHopProtocol=ipMRouteNextHopProtocol, PYSNMP_MODULE_ID=ipMRouteStdMIB, ipMRouteNextHopExpiryTime=ipMRouteNextHopExpiryTime, ipMRouteMIBHCInterfaceGroup=ipMRouteMIBHCInterfaceGroup, ipMRouteScopeNameDefault=ipMRouteScopeNameDefault, ipMRouteNextHopState=ipMRouteNextHopState, ipMRouteInterfaceInMcastOctets=ipMRouteInterfaceInMcastOctets, ipMRouteGroup=ipMRouteGroup, ipMRouteNextHopGroup=ipMRouteNextHopGroup, ipMRouteMIBGroups=ipMRouteMIBGroups, ipMRouteInterfaceIfIndex=ipMRouteInterfaceIfIndex, ipMRouteNextHopTable=ipMRouteNextHopTable, ipMRouteInterfaceTtl=ipMRouteInterfaceTtl, ipMRouteScopeNameString=ipMRouteScopeNameString, ipMRouteSource=ipMRouteSource, ipMRouteScopeNameStatus=ipMRouteScopeNameStatus, ipMRouteScopeNameAddressMask=ipMRouteScopeNameAddressMask, ipMRouteRtType=ipMRouteRtType, ipMRouteNextHopIfIndex=ipMRouteNextHopIfIndex, ipMRouteInterfaceOutMcastOctets=ipMRouteInterfaceOutMcastOctets, ipMRouteMIBHopCountGroup=ipMRouteMIBHopCountGroup, ipMRouteBoundaryStatus=ipMRouteBoundaryStatus, ipMRouteNextHopSourceMask=ipMRouteNextHopSourceMask, ipMRouteEnable=ipMRouteEnable, ipMRouteEntry=ipMRouteEntry, ipMRouteScopeNameTable=ipMRouteScopeNameTable, ipMRouteTable=ipMRouteTable, ipMRouteRtMask=ipMRouteRtMask)



"""Functionality to parse and validate the legacy polldevs.cf config file"""

from typing import Iterator, TextIO, Tuple

from pydantic import ValidationError

from zino.config.models import PollDevice


def read_polldevs(filename: str) -> Tuple[dict[str, PollDevice], dict[str, str]]:
    """
    Reads and parses the legacy `polldevs.cf` format, returning a dictionary of device names and the associated
    PollDevice object and a dictionary of default settings

    This parser is slightly more lax than the original Tcl-based parser, in that it allows multiple empty lines or
    multiple spaces in value assignments.
    """
    defaults = {}
    devices = {}
    try:
        with open(filename, "r") as devs:
            for lineno, section in _read_conf_sections(devs):
                if _contains_defaults(section):
                    defaults.update(_parse_defaults(section))
                    continue

                try:
                    device = PollDevice(**(defaults | section))
                    devices[device.name] = device
                except ValidationError as error:
                    first_error = error.errors()[0]
                    device_name = section.get("name", "N/A")
                    attribute = first_error["loc"][0]
                    raise InvalidConfiguration(
                        f"Validation error in device block {device_name!r}: {first_error['msg']} ({attribute!r})",
                        filename=filename,
                        lineno=lineno,
                    )

    except InvalidConfiguration as error:
        error.filename = filename
        raise

    return devices, defaults


def _read_conf_sections(filehandle: TextIO) -> Iterator[Tuple[int, dict]]:
    """Reads and yields individual configuration sections from `polldevs.cf`.

    Each yielded value is a two-tuple of the first line number of the section and the parsed section as a dict.
    """
    section = {}
    first_line = None
    for lineno, line in enumerate(filehandle):
        if not first_line:
            first_line = lineno + 1
        line = line.strip()
        if line.startswith("#"):
            continue
        if not line:
            if section:
                yield first_line, section
                section = {}
                first_line = None
            continue
        try:
            key, value = line.split(":", maxsplit=1)
        except ValueError:
            raise InvalidConfiguration(f"{line!r} is not a valid configuration line", lineno=lineno + 1)
        section[key.strip()] = value.strip()
    if section:
        yield first_line, section


def _contains_defaults(section: dict) -> bool:
    return any(key.startswith("default ") for key in section)


def _parse_defaults(section: dict) -> dict:
    return {key.split(" ", maxsplit=1)[1]: value for key, value in section.items() if key.startswith("default ")}


class InvalidConfiguration(Exception):
    def __init__(self, message=None, filename=None, lineno=None):
        self.filename = filename
        self.lineno = lineno
        super().__init__(message)

    def __str__(self):
        location = [item for item in (self.filename, self.lineno) if item]
        if location:
            location = ":".join(str(item) for item in location)
            return f"{location}: {super().__str__()}"
        return super().__str__()



#
# PySNMP MIB module CISCO-CONFIG-MAN-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file:///home/mvold/d/zino/mibs/CISCO-CONFIG-MAN-MIB.my
# Produced by pysmi-0.3.4 at Thu Jul 18 11:45:22 2024
# On host agrajag platform Linux version 6.6.37 by user mvold
# Using Python version 3.11.9 (main, Apr  2 2024, 08:25:04) [GCC 13.2.0]
#
OctetString, Integer, ObjectIdentifier = mibBuilder.importSymbols("ASN1", "OctetString", "Integer", "ObjectIdentifier")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsUnion, ValueSizeConstraint, ValueRangeConstraint, ConstraintsIntersection, SingleValueConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsUnion", "ValueSizeConstraint", "ValueRangeConstraint", "ConstraintsIntersection", "SingleValueConstraint")
ciscoMgmt, = mibBuilder.importSymbols("CISCO-SMI", "ciscoMgmt")
ModuleCompliance, NotificationGroup, ObjectGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ModuleCompliance", "NotificationGroup", "ObjectGroup")
TimeTicks, Counter32, Gauge32, ObjectIdentity, Unsigned32, MibScalar, MibTable, MibTableRow, MibTableColumn, Integer32, Bits, Counter64, NotificationType, MibIdentifier, IpAddress, ModuleIdentity, iso = mibBuilder.importSymbols("SNMPv2-SMI", "TimeTicks", "Counter32", "Gauge32", "ObjectIdentity", "Unsigned32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "Integer32", "Bits", "Counter64", "NotificationType", "MibIdentifier", "IpAddress", "ModuleIdentity", "iso")
TextualConvention, DisplayString = mibBuilder.importSymbols("SNMPv2-TC", "TextualConvention", "DisplayString")
ciscoConfigManMIB = ModuleIdentity((1, 3, 6, 1, 4, 1, 9, 9, 43))
ciscoConfigManMIB.setRevisions(('1995-11-28 00:00',))
if mibBuilder.loadTexts: ciscoConfigManMIB.setLastUpdated('9511280000Z')
if mibBuilder.loadTexts: ciscoConfigManMIB.setOrganization('Cisco Systems, Inc.')
ciscoConfigManMIBObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 43, 1))
ccmHistory = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1))
class HistoryEventMedium(TextualConvention, Integer32):
    status = 'current'
    subtypeSpec = Integer32.subtypeSpec + ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4, 5, 6, 7))
    namedValues = NamedValues(("erase", 1), ("commandSource", 2), ("running", 3), ("startup", 4), ("local", 5), ("networkTftp", 6), ("networkRcp", 7))

ccmHistoryRunningLastChanged = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 1), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryRunningLastChanged.setStatus('current')
ccmHistoryRunningLastSaved = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 2), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryRunningLastSaved.setStatus('current')
ccmHistoryStartupLastChanged = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 3), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryStartupLastChanged.setStatus('current')
ccmHistoryMaxEventEntries = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 4), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 2147483647))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryMaxEventEntries.setStatus('current')
ccmHistoryEventEntriesBumped = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 5), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventEntriesBumped.setStatus('current')
ccmHistoryEventTable = MibTable((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6), )
if mibBuilder.loadTexts: ccmHistoryEventTable.setStatus('current')
ccmHistoryEventEntry = MibTableRow((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1), ).setIndexNames((0, "CISCO-CONFIG-MAN-MIB", "ccmHistoryEventIndex"))
if mibBuilder.loadTexts: ccmHistoryEventEntry.setStatus('current')
ccmHistoryEventIndex = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 1), Integer32().subtype(subtypeSpec=ValueRangeConstraint(1, 2147483647)))
if mibBuilder.loadTexts: ccmHistoryEventIndex.setStatus('current')
ccmHistoryEventTime = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 2), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventTime.setStatus('current')
ccmHistoryEventCommandSource = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 3), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("commandLine", 1), ("snmp", 2)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventCommandSource.setStatus('current')
ccmHistoryEventConfigSource = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 4), HistoryEventMedium()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventConfigSource.setStatus('current')
ccmHistoryEventConfigDestination = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 5), HistoryEventMedium()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventConfigDestination.setStatus('current')
ccmHistoryEventTerminalType = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 6), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4, 5, 6))).clone(namedValues=NamedValues(("notApplicable", 1), ("unknown", 2), ("console", 3), ("terminal", 4), ("virtual", 5), ("auxiliary", 6)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventTerminalType.setStatus('current')
ccmHistoryEventTerminalNumber = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 7), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventTerminalNumber.setStatus('current')
ccmHistoryEventTerminalUser = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 8), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 64))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventTerminalUser.setStatus('current')
ccmHistoryEventTerminalLocation = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 9), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 64))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventTerminalLocation.setStatus('current')
ccmHistoryEventCommandSourceAddress = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 10), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventCommandSourceAddress.setStatus('current')
ccmHistoryEventVirtualHostName = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 11), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 64))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventVirtualHostName.setStatus('current')
ccmHistoryEventServerAddress = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 12), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventServerAddress.setStatus('current')
ccmHistoryEventFile = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 13), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 64))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventFile.setStatus('current')
ccmHistoryEventRcpUser = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 9, 43, 1, 1, 6, 1, 14), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 64))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ccmHistoryEventRcpUser.setStatus('current')
ciscoConfigManMIBNotificationPrefix = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 43, 2))
ciscoConfigManMIBNotifications = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 43, 2, 0))
ciscoConfigManEvent = NotificationType((1, 3, 6, 1, 4, 1, 9, 9, 43, 2, 0, 1)).setObjects(("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventCommandSource"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventConfigSource"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventConfigDestination"))
if mibBuilder.loadTexts: ciscoConfigManEvent.setStatus('current')
ciscoConfigManMIBConformance = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 43, 3))
ciscoConfigManMIBCompliances = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 43, 3, 1))
ciscoConfigManMIBGroups = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 43, 3, 2))
ciscoConfigManMIBCompliance = ModuleCompliance((1, 3, 6, 1, 4, 1, 9, 9, 43, 3, 1, 1)).setObjects(("CISCO-CONFIG-MAN-MIB", "ciscoConfigManHistoryGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ciscoConfigManMIBCompliance = ciscoConfigManMIBCompliance.setStatus('current')
ciscoConfigManHistoryGroup = ObjectGroup((1, 3, 6, 1, 4, 1, 9, 9, 43, 3, 2, 1)).setObjects(("CISCO-CONFIG-MAN-MIB", "ccmHistoryRunningLastChanged"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryRunningLastSaved"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryStartupLastChanged"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryMaxEventEntries"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventEntriesBumped"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventTime"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventCommandSource"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventConfigSource"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventConfigDestination"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventTerminalType"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventTerminalNumber"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventTerminalUser"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventTerminalLocation"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventCommandSourceAddress"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventVirtualHostName"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventServerAddress"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventFile"), ("CISCO-CONFIG-MAN-MIB", "ccmHistoryEventRcpUser"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ciscoConfigManHistoryGroup = ciscoConfigManHistoryGroup.setStatus('current')
mibBuilder.exportSymbols("CISCO-CONFIG-MAN-MIB", ccmHistoryEventEntry=ccmHistoryEventEntry, ccmHistoryRunningLastSaved=ccmHistoryRunningLastSaved, HistoryEventMedium=HistoryEventMedium, ccmHistoryEventConfigDestination=ccmHistoryEventConfigDestination, ccmHistoryEventServerAddress=ccmHistoryEventServerAddress, ccmHistoryEventTerminalLocation=ccmHistoryEventTerminalLocation, ciscoConfigManMIBCompliance=ciscoConfigManMIBCompliance, ccmHistoryEventTerminalNumber=ccmHistoryEventTerminalNumber, ccmHistoryEventTerminalType=ccmHistoryEventTerminalType, ccmHistoryEventCommandSourceAddress=ccmHistoryEventCommandSourceAddress, PYSNMP_MODULE_ID=ciscoConfigManMIB, ccmHistoryEventConfigSource=ccmHistoryEventConfigSource, ccmHistoryEventFile=ccmHistoryEventFile, ciscoConfigManHistoryGroup=ciscoConfigManHistoryGroup, ccmHistoryEventVirtualHostName=ccmHistoryEventVirtualHostName, ccmHistoryStartupLastChanged=ccmHistoryStartupLastChanged, ccmHistory=ccmHistory, ciscoConfigManMIBConformance=ciscoConfigManMIBConformance, ciscoConfigManMIBNotificationPrefix=ciscoConfigManMIBNotificationPrefix, ciscoConfigManMIBGroups=ciscoConfigManMIBGroups, ciscoConfigManMIB=ciscoConfigManMIB, ciscoConfigManMIBObjects=ciscoConfigManMIBObjects, ciscoConfigManMIBNotifications=ciscoConfigManMIBNotifications, ccmHistoryEventTerminalUser=ccmHistoryEventTerminalUser, ccmHistoryEventIndex=ccmHistoryEventIndex, ccmHistoryEventEntriesBumped=ccmHistoryEventEntriesBumped, ccmHistoryEventCommandSource=ccmHistoryEventCommandSource, ccmHistoryRunningLastChanged=ccmHistoryRunningLastChanged, ccmHistoryMaxEventEntries=ccmHistoryMaxEventEntries, ccmHistoryEventRcpUser=ccmHistoryEventRcpUser, ccmHistoryEventTable=ccmHistoryEventTable, ciscoConfigManMIBCompliances=ciscoConfigManMIBCompliances, ccmHistoryEventTime=ccmHistoryEventTime, ciscoConfigManEvent=ciscoConfigManEvent)



import asyncio

from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity import config, engine
from pysnmp.entity.rfc3413 import ntfrcv

# from pysnmp.hlapi.asyncio import UdpTransportTarget, Udp6TransportTarget


stop = asyncio.Event()


def main():
    try:
        asyncio.run(mofo())
    except (KeyboardInterrupt, SystemExit):
        stop.set()


async def mofo():
    await setup_trapd()
    await stop.wait()


async def setup_trapd():
    loop = asyncio.get_event_loop()
    # await loop.create_datagram_endpoint(lambda: EchoProtocol(), ("127.0.0.1", 1162))
    # Create SNMP engine with autogenernated engineID and pre-bound
    # to socket transport dispatcher
    snmpEngine = engine.SnmpEngine()

    # Transport setup

    # UDP over IPv4, first listening interface/port
    transport = udp.UdpTransport(loop=asyncio.get_event_loop()).openServerMode(("127.0.0.1", 1162))
    # This attribute needs to be awaited to ensure the socket is really opened,
    # unsure of why PySNMP doesn't do this, or how else it's supposed to be achieved
    await transport._lport
    config.addTransport(snmpEngine, udp.domainName + (1,), transport)
    print("transport added")
    config.addV1System(snmpEngine, "one", "public")
    config.addV1System(snmpEngine, "two", "public2")
    # Specify security settings per SecurityName (SNMPv1 - 0, SNMPv2c - 1)
    # config.addTargetParams(snmpEngine, 'my-creds', 'my-area', 'noAuthNoPriv', 1)

    # Register SNMP Application at the SNMP engine
    ntfrcv.NotificationReceiver(snmpEngine, callback)
    snmpEngine.transportDispatcher.jobStarted(1)  # this job would never finish
    # try:
    #     snmpEngine.transportDispatcher.runDispatcher()
    #
    # finally:
    #     snmpEngine.transportDispatcher.closeDispatcher()

    loop = asyncio.get_event_loop()


def callback(snmpEngine, stateReference, contextEngineId, contextName, varBinds, cbCtx):
    print(
        'Notification from ContextEngineId "%s", ContextName "%s"'
        % (contextEngineId.prettyPrint(), contextName.prettyPrint())
    )
    for name, val in varBinds:
        print("%s = %s" % (name.prettyPrint(), val.prettyPrint()))


class EchoProtocol(asyncio.DatagramProtocol):
    def datagram_received(self, data, addr):
        message = data.decode()
        print(f"Received {message} from {addr}")


if __name__ == "__main__":
    main()



import logging

from zino.snmp.base import NoSuchNameError, VarBindError
from zino.statemodels import AlarmEvent, AlarmType
from zino.tasks.task import Task

_logger = logging.getLogger(__name__)


class JuniperAlarmTask(Task):
    """Checks yellow and red alarm count for juniper device."""

    async def run(self):
        if not self.device_state.is_juniper:
            return

        try:
            yellow_alarm_count, red_alarm_count = await self._get_juniper_alarms()
        except (NoSuchNameError, TypeError, VarBindError):
            return

        if not self.device_state.alarms:
            self.device_state.alarms = {
                "yellow": 0,
                "red": 0,
            }

        if self.device_state.alarms["yellow"] != yellow_alarm_count:
            self.create_alarm_event(
                color="yellow", old_alarm_count=self.device_state.alarms["yellow"], new_alarm_count=yellow_alarm_count
            )
            self.device_state.alarms["yellow"] = yellow_alarm_count

        if self.device_state.alarms["red"] != red_alarm_count:
            self.create_alarm_event(
                color="red", old_alarm_count=self.device_state.alarms["red"], new_alarm_count=red_alarm_count
            )
            self.device_state.alarms["red"] = red_alarm_count

    async def _get_juniper_alarms(self):
        yellow_alarm_count = await self.snmp.get("JUNIPER-ALARM-MIB", "jnxYellowAlarmCount", 0)
        red_alarm_count = await self.snmp.get("JUNIPER-ALARM-MIB", "jnxRedAlarmCount", 0)
        if yellow_alarm_count:
            yellow_alarm_count = yellow_alarm_count.value
        if red_alarm_count:
            red_alarm_count = red_alarm_count.value

        if not isinstance(yellow_alarm_count, int) or not isinstance(red_alarm_count, int):
            _logger.error(
                "Device %s returns alarm count not of type int. Yellow alarm count: type %s. Red alarm count: type %s.",
                self.device.name,
                type(yellow_alarm_count),
                type(red_alarm_count),
            )
            _logger.debug(
                "Yellow alarm count: value %r. Red alarm count: value %r.",
                yellow_alarm_count,
                red_alarm_count,
            )
            raise TypeError

        return yellow_alarm_count, red_alarm_count

    def create_alarm_event(self, color: AlarmType, old_alarm_count: int, new_alarm_count: int):
        alarm_event = self.state.events.get_or_create_event(
            device_name=self.device.name,
            subindex=color,
            event_class=AlarmEvent,
        )

        log = f"alarms went from {old_alarm_count} to {new_alarm_count}"
        alarm_event.alarm_type = color
        alarm_event.alarm_count = new_alarm_count
        alarm_event.add_log(f"{self.device.name} {color} {log}")
        alarm_event.polladdr = self.device.address
        alarm_event.priority = self.device.priority
        alarm_event.lastevent = log
        self.state.events.commit(alarm_event)



"""Common fixtures for trap tests"""

import ipaddress

import pytest

from zino.statemodels import DeviceState
from zino.trapd.base import TrapOriginator


@pytest.fixture
def localhost_trap_originator():
    addr = ipaddress.IPv4Address("127.0.0.1")
    device = DeviceState(name="localhost", addresses=set((addr,)))
    return TrapOriginator(address=addr, port=162, device=device)



import logging
from datetime import datetime, timedelta, timezone

from zino.state import ZinoState
from zino.stateconverter.linedata import LineData
from zino.stateconverter.utils import OldState, parse_ip, parse_log_and_history
from zino.statemodels import (
    AlarmEvent,
    BFDEvent,
    BFDSessState,
    BGPAdminStatus,
    BGPEvent,
    BGPOperState,
    EventState,
    FlapState,
    InterfaceState,
    PortStateEvent,
    ReachabilityEvent,
    ReachabilityState,
)

_log = logging.getLogger(__name__)


event_name_to_type = {
    "bgp": BGPEvent,
    "bfd": BFDEvent,
    "reachability": ReachabilityEvent,
    "alarm": AlarmEvent,
    "portstate": PortStateEvent,
}


def set_event_state(
    old_state: OldState,
    new_state: ZinoState,
):
    id_to_type = {}

    # Register event type per id
    for linedata in old_state["::EventAttrs"]:
        if linedata.identifiers[0] == "type":
            id_to_type[int(linedata.identifiers[1])] = linedata.value

    for linedata in old_state["::EventAttrs"]:
        try:
            _set_event_attrs(linedata, new_state, id_to_type)
        except ValueError as e:
            _log.error(f"Error setting event attribute: {str(e)}")
    new_state.events._rebuild_indexes()


def _set_event_attrs(linedata: LineData, state: ZinoState, id_to_type: dict[int, str]):
    event_field = linedata.identifiers[0]
    event_id = int(linedata.identifiers[1])
    state.events.last_event_id = max(state.events.last_event_id, event_id)
    if event_id in state.events.events:
        event = state.events.events[event_id]
    else:
        event_type = id_to_type[event_id]
        event_class = event_name_to_type[event_type]
        # router needs to be set now since its a required field, will be overwritten later
        event = event_class(id=event_id, router="placeholder")
    if event_field == "priority":
        event.priority = int(linedata.value)
    elif event_field == "history":
        event.history = parse_log_and_history(linedata.value)
    elif event_field == "bgpOS":
        event.operational_state = BGPOperState(linedata.value)
    elif event_field == "bgpAS":
        event.admin_status = BGPAdminStatus(linedata.value)
    elif event_field == "lastevent":
        event.lastevent = linedata.value
    elif event_field == "log":
        event.log = parse_log_and_history(linedata.value)
    elif event_field == "polladdr":
        event.polladdr = parse_ip(linedata.value)
    elif event_field == "opened":
        event.opened = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    elif event_field == "peer-uptime":
        event.peer_uptime = int(linedata.value)
    elif event_field == "remote-AS":
        event.remote_as = int(linedata.value)
    elif event_field == "remote-addr":
        event.remote_address = parse_ip(linedata.value)
    elif event_field == "router":
        event.router = linedata.value
    elif event_field == "state":
        event.state = EventState(linedata.value)
    elif event_field == "updated":
        event.updated = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    elif event_field == "ac-down":
        event.ac_down = timedelta(seconds=int(linedata.value))
    elif event_field == "descr":
        event.descr = linedata.value
    elif event_field == "flaps":
        event.flaps = int(linedata.value)
    elif event_field == "flapstate":
        event.flapstate = FlapState(linedata.value)
    elif event_field == "ifindex":
        event.ifindex = int(linedata.value)
    elif event_field == "portstate":
        event.portstate = InterfaceState(linedata.value)
    elif event_field == "port":
        event.port = linedata.value
    elif event_field == "bfdAddr":
        if "unknown" in linedata.value:
            pass
        else:
            event.bfdaddr = parse_ip(linedata.value)
    elif event_field == "bfdDiscr":
        event.bfddiscr = int(linedata.value)
    elif event_field == "bfdIx":
        event.bfdix = int(linedata.value)
    elif event_field == "bfdState":
        event.bfdstate = BFDSessState(linedata.value)
    elif event_field == "lasttrans":
        event.lasttrans = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    elif event_field == "alarm-count":
        event.alarm_count = int(linedata.value)
    elif event_field == "alarm-type":
        event.alarm_type = linedata.value
    elif event_field == "Neigh-rDNS":
        event.neigh_rdns = linedata.value
    elif event_field == "reachability":
        event.reachability = ReachabilityState(linedata.value)
    elif event_field == "reason":
        event.reason = linedata.value
    elif event_field in ["id", "type"]:
        # These are set via other means
        pass
    else:
        raise ValueError(f"Unknown event attribute {event_field}")
    state.events.events[event.id] = event



"""SNMP trap 'daemon' for Zino 2 implemented using netsnmp-cffi"""

import asyncio
import logging
from ipaddress import ip_address
from typing import Dict, Optional

from netsnmpy.netsnmp import SNMPVariable
from netsnmpy.trapsession import SNMPTrap, SNMPTrapSession

import zino.state
from zino.config.models import PollDevice
from zino.snmp.netsnmpy_backend import _convert_snmp_variable
from zino.trapd.base import (
    TrapMessage,
    TrapOriginator,
    TrapReceiverBase,
    TrapVarBind,
)

_logger = logging.getLogger(__name__)


class TrapReceiver(TrapReceiverBase):
    """Zino Adapter for SNMP trap reception using netsnmp-cffi.

    Zino 1 accepts traps with any community string, as long as its origin is any one of the devices configured in
    the pollfile.  The PySNMP back-end will only accept traps with one of the configured community strings.  *This*
    back-end, however, will accept and pass on traps with any community string until `add_community()` is called
    to configure at least one filter.
    """

    def __init__(
        self,
        address: str = "0.0.0.0",
        port: int = 162,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        state: Optional[zino.state.ZinoState] = None,
        polldevs: Optional[Dict[str, PollDevice]] = None,
    ):
        super().__init__(address, port, loop, state, polldevs)
        self._session = SNMPTrapSession(ip_address(address), port)

    async def open(self):
        """Opens the UDP transport socket and starts receiving traps"""
        self._session.add_observer(self.trap_received)
        self._session.open()
        _logger.info("Listening for incoming SNMP traps on %r", (self.address, self.port))

    def close(self):
        """Closes the running SNMP engine and its associated ports"""
        self._session.close()

    def trap_received(self, trap: SNMPTrap):
        """Callback function that receives all trap messages from the Net-SNMP transport"""
        router = self._lookup_device(trap.source)
        # netsnmpy doesn't currently provide the source port:
        origin = TrapOriginator(address=trap.source, port=None, device=router)
        if not self._verify_trap(trap, origin):
            return

        _logger.debug("Trap from %s (%s)", router.name, trap.source)

        zino_trap = TrapMessage(agent=origin)
        for variable in trap.variables:
            _logger.debug("%s", variable)
            try:
                identifier, value = _convert_snmp_variable(variable)
            except ValueError:
                _logger.debug(
                    "Could not resolve SNMP variable %s: %s (Maybe MIB not loaded?)", variable, variable.value
                )
                return

            # This should really be part of netsnmp-cffi, but it isn't currently
            raw_value = value = variable.value
            enum_value = variable.enum_value
            if enum_value:
                value = enum_value
            elif variable.textual_convention == "DisplayString":
                value = variable.value.decode("utf-8")

            zino_trap.variables.append(
                TrapVarBind(variable.oid, identifier.mib, identifier.object, identifier.index, raw_value, value)
            )

        # TODO do some time calculations, but ask Håvard what the deal is with RestartTime vs. BootTime

        if trap.trap_oid:
            try:
                trap_identifier, _ = _convert_snmp_variable(SNMPVariable(trap.trap_oid, None))
            except ValueError:
                _logger.error("Could not resolve trap OID %s (Maybe MIB not loaded?)", trap.trap_oid)
                return
            zino_trap.mib, zino_trap.name = trap_identifier.mib, trap_identifier.object
            _logger.debug("Trap from %s identified as %r", trap.source, trap_identifier)
        asyncio.ensure_future(self.dispatch_trap(zino_trap))

    def _verify_trap(self, netsnmp_trap: SNMPTrap, origin: TrapOriginator) -> bool:
        if not origin.device:
            _logger.debug("ignored trap from %s (not a box we monitor?)", origin.address)
            return False

        source_name = origin.device.name or origin.address

        if self._communities and netsnmp_trap.community not in self._communities:
            _logger.info("Trap from %s with unknown community string %r, ignoring", source_name, netsnmp_trap.community)
            return False

        if not netsnmp_trap.trap_oid:
            _logger.info("Trap from %s did not contain a snmpTrapOID value, ignoring", source_name)
            return False

        if not netsnmp_trap.uptime:
            _logger.info("Trap from %s did not contain a sysUpTime value, ignoring", source_name)
            return False

        return True



"""This module implements link trap handling"""

import logging
import re
from datetime import datetime, timedelta
from typing import Optional

import zino.state
import zino.time
from zino.flaps import PortIndex
from zino.statemodels import (
    DeviceState,
    FlapState,
    InterfaceState,
    Port,
    PortStateEvent,
)
from zino.tasks.linkstatetask import LinkStateTask
from zino.trapd.base import TrapMessage, TrapObserver

_logger = logging.getLogger(__name__)
TRAP_WINDOW = timedelta(minutes=5)
IMMEDIATELY = timedelta(seconds=0)
FIRST_REVERIFICATION = IMMEDIATELY
SECOND_REVERIFICATION = timedelta(minutes=2)


class LinkTrapObserver(TrapObserver):
    WANTED_TRAPS = {
        ("IF-MIB", "linkUp"),
        ("IF-MIB", "linkDown"),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._last_same_trap: dict[PortIndex, datetime] = {}

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        _logger.debug("%s: %s (vars: %s)", trap.agent.device.name, trap.name, ", ".join(v.var for v in trap.variables))

        if "ifIndex" in trap:
            ifindex = trap.get_all("ifIndex")[0].value
        else:
            _logger.warning("%s: %s trap contained no ifIndex value, ignoring", trap.agent.device.name, trap.name)
            return False
        port = trap.agent.device.ports.get(ifindex) if ifindex > 0 else None
        if not port:
            _logger.warning(
                "%s: %s trap referenced unknown port (ix %s), ignoring", trap.agent.device.name, trap.name, ifindex
            )
            return False

        # The legacy Zino also looks for `locIfDescr` in the trap at this point, but this is an ancient
        # Cisco-specific variable we no longer see.  It might have been replaced by `cieIfOperStatusCause`,
        # but we have no more Cisco devices to test on, so this has been deliberately left out.

        is_up = trap.name == "linkUp"
        self.handle_link_transition(trap.agent.device, port, is_up)

    def handle_link_transition(self, device: DeviceState, port: Port, is_up: bool, reason: Optional[str] = None):
        """This method is called when a linkUp or linkDown trap is received, and updates event state accordingly -
        including the logic for handling flapping states.
        """
        if self.is_port_ignored_by_patterns(device, port.ifdescr):
            return

        # At this point, legacy Zino would ignore so-called "insignificant" interfaces.  As mentioned in the
        # LinkStateTask docstring, this concept has deliberately not been ported to Zino 2.

        _logger.debug(
            "%s: intf %s ix %d link%s%s",
            device.name,
            port.ifdescr,
            port.ifindex,
            "Up" if is_up else "Down",
            f", {reason}" if reason else "",
        )

        if self.is_port_ignored_by_policy(device, port, is_up):
            return

        index = (device.name, port.ifindex)
        flap = self.state.flapping.update_interface_flap(index)

        new_state = InterfaceState.UP if is_up else InterfaceState.DOWN

        if flap.is_flapping():
            if not flap.in_active_flap_state:
                event = self.update_flapping_event(device, index, new_state, port, reason)
            else:
                event = self.state.events.get(device.name, port.ifindex, PortStateEvent)

            if event:
                # Update the event flap count, but explicitly DO NOT commit the change here, as committing during
                # flap storms would cause a notification storm as well
                event.flaps = flap.flaps

            if flap.flaps % 100 == 0:
                _logger.info(
                    '%s: intf "%s" ix %d (%s), %d flaps, penalty %5.2f',
                    device.name,
                    port.ifdescr,
                    port.ifindex,
                    port.ifalias,
                    flap.flaps,
                    flap.hist_val,
                )
        else:
            self.update_non_flapping_event(device, index, new_state, port, reason)

    def update_flapping_event(
        self, device: DeviceState, interface: PortIndex, new_state: InterfaceState, port: Port, reason: Optional[str]
    ) -> PortStateEvent:
        """Opens or returns an existing portstate event for a port that was not previously known to be flapping.

        It also updates the flap state of the port to indicate it is currently in a flapping state.
        """
        event = self.open_portstate_event(device, new_state, port, reason)

        flap = self.state.flapping.interfaces[interface]
        event.flapstate = FlapState.FLAPPING
        event.flaps = flap.flaps
        msg = (
            f'{device.name}: intf "{port.ifdescr}" ix {port.ifindex} ({port.ifalias}) flapping, '
            f"{event.flaps} flaps, penalty {flap.hist_val:5.2f}"
        )
        _logger.info(msg)
        event.add_log(msg)
        self.state.events.commit(event)
        flap.in_active_flap_state = True
        return event

    def update_non_flapping_event(
        self, device: DeviceState, interface: PortIndex, new_state: InterfaceState, port: Port, reason: Optional[str]
    ) -> PortStateEvent:
        """Opens or returns an existing portstate event for a port that isn't currently flapping.

        It also removes the recorded flap stats for the port if was previously known to be in a flapping state.
        """
        event = self.open_portstate_event(device, new_state, port, reason)

        if flap := self.state.flapping.interfaces.get(interface):
            event.flaps = flap.flaps
            if flap.in_active_flap_state:
                event.flapstate = FlapState.STABLE
                msg = f'{device.name}: intf "{port.ifdescr}" ix {port.ifindex} ({port.ifalias}) stopped flapping'
                _logger.info(msg)
                event.add_log(msg)
                self.state.flapping.unflap(interface)

        msg = f'{device.name}: intf "{port.ifdescr}" ix {port.ifindex} link{new_state.capitalize()}'
        if reason:
            msg = f'{msg}, "{reason}"'
        _logger.info(msg)
        event.add_log(msg)
        self.state.events.commit(event)

        if polldev := self.polldevs.get(device.name):
            poll = LinkStateTask(device=polldev, state=self.state, config=zino.state.config)
            poll.schedule_verification_of_single_port(port.ifindex, deadline=FIRST_REVERIFICATION, reason="trap-verify")
            poll.schedule_verification_of_single_port(
                port.ifindex, deadline=SECOND_REVERIFICATION, reason="trap-verify-2"
            )
        else:
            _logger.warning("No polldev config found for %s", device.name)

        return event

    def open_portstate_event(
        self, device: DeviceState, new_state: InterfaceState, port: Port, reason: Optional[str]
    ) -> PortStateEvent:
        """Opens or returns an open port state event for the port referenced in an incoming trap message"""
        event = self.state.events.get_or_create_event(device.name, port.ifindex, PortStateEvent)
        event.portstate = new_state
        event.port = port.ifdescr
        event.ifindex = port.ifindex
        port.state = new_state
        if polldev := self.polldevs.get(device.name):
            event.polladdr = polldev.address
            event.priority = polldev.priority
        event.descr = port.ifalias
        if reason:
            event.reason = reason
        return event

    def is_port_ignored_by_patterns(self, device: DeviceState, ifdescr: str) -> bool:
        if watch_pattern := self.get_watch_pattern(device):
            if not re.search(watch_pattern, ifdescr):
                return True

        if ignore_pattern := self.get_ignore_pattern(device):
            if re.search(ignore_pattern, ifdescr):
                return True

        return False

    def is_port_ignored_by_policy(self, device: DeviceState, port: Port, is_up: bool) -> bool:
        """Verifies that a link trap should be ignored, according to internal policies.

        As commented in the original Zino 1 implementation:

        > Should we ignore this trap message?
        > If there's an open case for this router port, no.
        > If received right after reload, yes.
        > If state reported is the same as the one we have recorded,
        > and a new trap does not reoccur within 5 minutes, yes.
        > Otherwise, don't ignore the trap message.
        """
        now = zino.time.now()

        # If there is an open case for router/port, do not ignore
        if self.state.events.get(device.name, subindex=port.ifindex, event_class=PortStateEvent):
            return False

        # If we don't know that the device restarted recently, do not ignore
        if not device.boot_time:
            _logger.info("Oops! Do not know restart time for %s", device.name)
            return False

        if not port.state:
            port.state = InterfaceState.UNKNOWN

        if now - device.boot_time > TRAP_WINDOW:
            # Do not ignore traps indicating different state than known
            current_state = InterfaceState.UP if is_up else InterfaceState.DOWN
            if port.state != current_state:
                return False

            # Do not ignore successive traps within a TRAP_WINDOW interval indicating same state as previously known
            index = (device.name, port.ifindex)
            if last_same_trap := self._last_same_trap.get(index):
                if now - last_same_trap < TRAP_WINDOW:
                    self._last_same_trap[index] = now
                    return False
            self._last_same_trap[index] = now

        # Ignoring traps the first 5 minutes after restart
        _logger.info(
            "Ignored %s trap for %s ix %s (state %s), restarted %s",
            "Up" if is_up else "Down",
            device.name,
            port.ifindex,
            port.state,
            device.boot_time.astimezone(),
        )
        return True

    def get_watch_pattern(self, device: DeviceState) -> Optional[str]:
        if device.name not in self.polldevs:
            return None
        return self.polldevs[device.name].watchpat

    def get_ignore_pattern(self, device: DeviceState) -> Optional[str]:
        if device.name not in self.polldevs:
            return None
        return self.polldevs[device.name].ignorepat



import ipaddress
import logging
from unittest.mock import Mock

import pytest

from zino.config.models import PollDevice
from zino.oid import OID
from zino.trapd.base import TrapMessage
from zino.trapobservers.bfd_traps import BFDTrapObserver


class TestBFDTrapObserver:
    async def test_when_bfd_trap_is_received_it_should_poll_all_affected_sessions(
        self, bfd_session_down_trap, monkeypatch, event_loop
    ):
        device = bfd_session_down_trap.agent.device
        polldevs_dict = {device.name: PollDevice(name=device.name, address=ipaddress.IPv4Address("127.0.0.1"))}
        called_future = event_loop.create_future()
        called_future.set_result(None)
        bfdtask_run = Mock(return_value=called_future)
        monkeypatch.setattr("zino.tasks.bfdtask.BFDTask.run", bfdtask_run)

        observer = BFDTrapObserver(state=Mock(), polldevs=polldevs_dict)
        await observer.handle_trap(trap=bfd_session_down_trap)

        assert bfdtask_run.call_count == 4

    async def test_when_polldevs_config_is_missing_it_should_do_nothing(
        self, bfd_session_down_trap, monkeypatch, event_loop
    ):
        called_future = event_loop.create_future()
        called_future.set_result(None)
        bfdtask_run = Mock(return_value=called_future)
        monkeypatch.setattr("zino.tasks.bfdtask.BFDTask.run", bfdtask_run)

        observer = BFDTrapObserver(state=Mock(), polldevs={})
        await observer.handle_trap(trap=bfd_session_down_trap)

        assert bfdtask_run.call_count == 0

    async def test_when_malformed_bfd_trap_is_received_it_should_log_and_return(
        self,
        malformed_bfd_session_down_trap,
        monkeypatch,
        event_loop,
        caplog,
    ):
        device = malformed_bfd_session_down_trap.agent.device
        polldevs_dict = {device.name: PollDevice(name=device.name, address=ipaddress.IPv4Address("127.0.0.1"))}
        called_future = event_loop.create_future()
        called_future.set_result(None)
        bfdtask_run = Mock(return_value=called_future)
        monkeypatch.setattr("zino.tasks.bfdtask.BFDTask.run", bfdtask_run)

        observer = BFDTrapObserver(state=Mock(), polldevs=polldevs_dict)
        with caplog.at_level(logging.INFO):
            await observer.handle_trap(trap=malformed_bfd_session_down_trap)
            assert "malformed" in caplog.text

        assert bfdtask_run.call_count == 0


@pytest.fixture
def bfd_session_down_trap(localhost_trap_originator) -> TrapMessage:
    """Returns a correct BFD session-down trap for 4 sessions, with internal state to match"""
    trap = TrapMessage(agent=localhost_trap_originator, mib="BGP4-V2-MIB-JUNIPER", name="bfdSessDown")
    trap.variables = [
        Mock(var="bfdSessDiag", instance=OID(".42"), value="pathDown"),
        Mock(var="bfdSessDiag", instance=OID(".45"), value="pathDown"),
    ]
    return trap


@pytest.fixture
def malformed_bfd_session_down_trap(localhost_trap_originator) -> TrapMessage:
    """Returns a BFD session-down trap with only a single bfdSessDiag value"""
    trap = TrapMessage(agent=localhost_trap_originator, mib="BGP4-V2-MIB-JUNIPER", name="bfdSessDown")
    trap.variables = [
        Mock(var="bfdSessDiag", instance=OID(".42"), value="pathDown"),
    ]
    return trap



import pytest

from zino.api.auth import AuthenticationFailure, authenticate, get_challenge, read_users


class TestAuthenticate:
    def test_when_challenge_response_is_ok_should_return_true(self, secrets_file):
        good_response = "b3e9ba156949152c5c6f3334cf86333b1daabfe9"
        assert authenticate(
            "user1",
            response=good_response,
            challenge="a97fbebb3eef2dfcf167645229c0c5fa9e92e3da",
            secrets_file=secrets_file,
        )

    def test_when_challenge_response_is_bad_should_raise_error(self, secrets_file):
        bad_response = "aaa9ba156949152c5c6f3334cf86333b1daabbbb"
        with pytest.raises(AuthenticationFailure):
            assert authenticate(
                "user1",
                response=bad_response,
                challenge="a97fbebb3eef2dfcf167645229c0c5fa9e92e3da",
                secrets_file=secrets_file,
            )

    def test_when_user_does_not_exist_should_raise_error(self, secrets_file):
        with pytest.raises(AuthenticationFailure):
            assert authenticate(
                "nobody",
                response="doesntmatter",
                challenge="doesntmatter",
                secrets_file=secrets_file,
            )

    def test_when_challenge_is_empty_should_raise_error(self, secrets_file):
        """Can be removed when challenge-less auth is supported"""
        with pytest.raises(AuthenticationFailure):
            assert authenticate(
                "user1",
                response="doesntmatter",
                secrets_file=secrets_file,
            )


class TestGetChallenge:
    def test_should_return_unique_challenges(self):
        """Challenge strings are expected to be random.  Once in a million years this could fail because of the
        nature of randomness, but it should also protect against the implementation being changed into something that
        returns a constant value.
        """
        count = 10
        challenges = set(get_challenge() for _ in range(count))
        assert len(challenges) == 10, "same challenge was produced more than once"

    def test_should_return_at_least_40_characters(self):
        assert len(get_challenge()) >= 40


class TestReadUsers:
    def test_result_should_match_input_file(self, secrets_file):
        expected = {
            "user1": "3c55d3cdc19876dfc8c0bb49da8c927a0ddff26d",
            "user2": "eb6fb35f5fbba6a6e43d3c893ad7a77dc793ceba",
            "user3": "c8a0b250edb2eabd9616b7b05a46e0ad28226fc2",
        }
        users = read_users(secrets_file)
        assert users == expected

    def test_should_not_crash_on_empty_lines_in_input(self, secrets_file_littered_with_empty_lines):
        assert read_users(secrets_file_littered_with_empty_lines)



import datetime
from typing import Iterator
from unittest.mock import patch

import pytest

from zino.statemodels import (
    DeviceMaintenance,
    EventState,
    MatchType,
    PortStateEvent,
    ReachabilityEvent,
)


class TestStart:
    def test_should_set_state_of_matching_events_to_ignored(self, state, reachability_event, matching_device_pm):
        assert reachability_event.state is EventState.OPEN
        matching_device_pm.start(state)
        assert state.events.get(reachability_event.router, None, ReachabilityEvent).state is EventState.IGNORED

    def test_should_add_started_events_to_event_ids_list(self, state, reachability_event, matching_device_pm):
        assert reachability_event.id not in matching_device_pm.event_ids
        matching_device_pm.start(state)
        assert reachability_event.id in matching_device_pm.event_ids

    def test_should_not_change_state_of_non_matching_events(self, state, reachability_event, nonmatching_device_pm):
        assert reachability_event.state is EventState.OPEN
        nonmatching_device_pm.start(state)
        assert state.events.get(reachability_event.router, None, ReachabilityEvent).state is EventState.OPEN

    def test_should_not_change_state_of_events_that_are_not_reachability_or_alarm_events(
        self, state, portstate_event, matching_device_pm
    ):
        assert portstate_event.state is EventState.OPEN
        matching_device_pm.start(state)
        assert (
            state.events.get(portstate_event.router, portstate_event.subindex, PortStateEvent).state is EventState.OPEN
        )

    def test_should_create_matching_event_with_ignored_state_if_it_does_not_exist(
        self, state, device, matching_device_pm
    ):
        assert not state.events.get(device.name, None, ReachabilityEvent)
        matching_device_pm.start(state)
        created_event = state.events.get(device.name, None, ReachabilityEvent)
        assert created_event
        assert created_event.state is EventState.IGNORED


class TestMatchesEvent:
    def test_should_return_true_for_matching_event(self, state, reachability_event, matching_device_pm):
        assert matching_device_pm.matches_event(reachability_event, state)

    def test_should_return_false_for_non_matching_events(self, state, reachability_event, nonmatching_device_pm):
        assert not nonmatching_device_pm.matches_event(reachability_event, state)

    def test_should_return_false_for_events_that_are_not_reachability_or_alarm_events(
        self, state, portstate_event, matching_device_pm
    ):
        assert not matching_device_pm.matches_event(portstate_event, state)


class TestMatchesDevice:
    @pytest.mark.parametrize("device_pm", [MatchType.EXACT, MatchType.REGEXP, MatchType.STR], indirect=True)
    def test_should_return_true_if_device_name_matches_expression(self, device, device_pm):
        assert device_pm.matches_device(device)

    @pytest.mark.parametrize("device_pm", [MatchType.REGEXP], indirect=True)
    def test_should_return_true_if_device_name_matches_start_of_regexp_expression(self, device, device_pm):
        assert device_pm.matches_device(device)

    @pytest.mark.parametrize("device_pm", [MatchType.REGEXP], indirect=True)
    def test_should_return_true_if_device_name_matches_not_start_of_regexp_expression(self, device, device_pm):
        device.name = "blabla" + device.name
        assert device_pm.matches_device(device)

    @pytest.mark.parametrize("device_pm", [MatchType.EXACT, MatchType.REGEXP, MatchType.STR], indirect=True)
    def test_should_return_false_if_device_name_does_not_match_expression(self, device, device_pm):
        device.name = "wrong"
        assert not device_pm.matches_device(device)

    @pytest.mark.parametrize("device_pm", [MatchType.INTF_REGEXP], indirect=True)
    def test_should_return_false_if_invalid_match_type(self, device, device_pm):
        assert not device_pm.matches_device(device)


@pytest.fixture
def device_pm(request, device) -> DeviceMaintenance:
    return DeviceMaintenance(
        start_time=datetime.datetime.now() - datetime.timedelta(days=1),
        end_time=datetime.datetime.now() + datetime.timedelta(days=1),
        match_type=request.param,
        match_expression=device.name,
        match_device=None,
    )


@pytest.fixture
def matching_device_pm(device) -> Iterator[DeviceMaintenance]:
    with patch("zino.statemodels.DeviceMaintenance.matches_device") as mock:
        mock.return_value = True
        yield DeviceMaintenance(
            start_time=datetime.datetime.now() - datetime.timedelta(days=1),
            end_time=datetime.datetime.now() + datetime.timedelta(days=1),
            match_type=MatchType.STR,
            match_expression=device.name,
            match_device=None,
        )


@pytest.fixture
def nonmatching_device_pm(device) -> Iterator[DeviceMaintenance]:
    with patch("zino.statemodels.DeviceMaintenance.matches_device") as mock:
        mock.return_value = False
        yield DeviceMaintenance(
            start_time=datetime.datetime.now() - datetime.timedelta(days=1),
            end_time=datetime.datetime.now() + datetime.timedelta(days=1),
            match_type=MatchType.STR,
            match_expression=device.name,
            match_device=None,
        )



import asyncio
import os
from unittest.mock import Mock, patch

import pytest
from netsnmpy import netsnmp
from netsnmpy.netsnmp import EndOfMibView, NoSuchInstance, NoSuchObject, SNMPVariable
from netsnmpy.oids import OID

from zino.config.models import PollDevice
from zino.snmp.base import (
    EndOfMibViewError,
    Identifier,
    MibNotFoundError,
    NoSuchInstanceError,
    NoSuchNameError,
    NoSuchObjectError,
    SNMPBackendError,
    SNMPBackendVersionError,
)
from zino.snmp.netsnmpy_backend import SNMP, init_backend, resolve_symbol


@pytest.fixture(scope="session")
def snmp_client(snmpsim, snmp_test_port):
    device = PollDevice(name="buick.lab.example.org", address="127.0.0.1", port=snmp_test_port)
    with SNMP(device) as snmp:
        yield snmp


@pytest.fixture(scope="session")
def ipv6_snmp_client(snmpsim, snmp_test_port) -> SNMP:
    device = PollDevice(name="buick.lab.example.org", address="::1", port=snmp_test_port)
    return SNMP(device)


@pytest.fixture()
def unreachable_snmp_client():
    future = asyncio.Future()
    future.set_exception(TimeoutError("Mock timeout"))
    timeout_mock = Mock(return_value=future)
    with patch.multiple(
        "zino.snmp.netsnmpy_backend.SNMP",
        get=timeout_mock,
        getnext=timeout_mock,
        getnext2=timeout_mock,
        walk=timeout_mock,
        getbulk=timeout_mock,
        bulkwalk=timeout_mock,
        getbulk2=timeout_mock,
        sparsewalk=timeout_mock,
    ):
        device = PollDevice(name="nonexist", address="127.0.0.1", community="invalid", port=666)
        yield SNMP(device)


class TestInitBackend:
    def test_when_netsnmp_is_too_old_it_should_raise(self):
        with patch("zino.snmp.netsnmpy_backend.netsnmp.get_version", return_value=(5, 7, 0)):
            with pytest.raises(SNMPBackendVersionError):
                init_backend()

    def test_when_mibdirs_envvar_is_not_set_it_should_set_it(self):
        env = os.environ.copy()
        env.pop("MIBDIRS", None)
        with patch("os.environ", env):
            init_backend()
            assert "MIBDIRS" in os.environ

    def test_when_vendored_mib_cannot_be_resolved_it_should_raise(self):
        with patch("zino.snmp.netsnmpy_backend.netsnmp.symbol_to_oid", side_effect=ValueError("Mock error")):
            with pytest.raises(SNMPBackendError):
                init_backend()


class TestSNMPRequestsResponseTypes:
    async def test_get(self, snmp_client):
        response = await snmp_client.get("SNMPv2-MIB", "sysUpTime", 0)
        assert isinstance(response.oid, OID)
        assert isinstance(response.value, int)

    async def test_get2_should_return_symbolic_identifiers(self, snmp_client):
        response = await snmp_client.get2(("IF-MIB", "ifName", 1), ("IF-MIB", "ifAlias", 1))
        assert len(list(response)) == 2
        assert any(identifier == Identifier("IF-MIB", "ifName", OID(".1")) for identifier, _ in response)
        assert any(identifier == Identifier("IF-MIB", "ifAlias", OID(".1")) for identifier, _ in response)

    async def test_when_mib_is_unkown_get2_should_raise_mibnotfounderror(self, snmp_client):
        with pytest.raises(MibNotFoundError):
            await snmp_client.get2(("FOOBAR-MIB", "ifName", 1), ("FOOBAR-MIB", "ifAlias", 1))

    async def test_getnext(self, snmp_client):
        response = await snmp_client.getnext("SNMPv2-MIB", "sysUpTime")
        assert isinstance(response.oid, OID)
        assert isinstance(response.value, int)

    async def test_getnext2_should_return_symbolic_identifiers(self, snmp_client):
        response = await snmp_client.getnext2(("IF-MIB", "ifName", "1"), ("IF-MIB", "ifAlias", "1"))
        assert len(list(response)) == 2
        assert any(identifier == Identifier("IF-MIB", "ifName", OID(".2")) for identifier, _ in response)
        assert any(identifier == Identifier("IF-MIB", "ifAlias", OID(".2")) for identifier, _ in response)

    async def test_walk(self, snmp_client):
        response = await snmp_client.walk("SNMPv2-MIB", "sysUpTime")
        assert response
        for mib_object in response:
            assert isinstance(mib_object.oid, OID)
            assert isinstance(mib_object.value, int)

    async def test_getbulk(self, snmp_client):
        response = await snmp_client.getbulk("SNMPv2-MIB", "sysUpTime")
        assert response
        for mib_object in response:
            assert isinstance(mib_object.oid, OID)
            assert isinstance(mib_object.value, int)

    async def test_getbulk2_should_have_expected_response(self, snmp_client):
        variables = ("ifIndex", "ifDescr", "ifAlias")
        response = await snmp_client.getbulk2(*(("IF-MIB", var) for var in variables))
        assert response
        for var_binds in response:
            assert len(var_binds) == len(variables)
            for ident, value in var_binds:
                assert ident.object in variables

    async def test_bulkwalk(self, snmp_client):
        response = await snmp_client.bulkwalk("SNMPv2-MIB", "sysUpTime")
        assert response
        for mib_object in response:
            assert isinstance(mib_object.oid, OID)
            assert isinstance(mib_object.value, int)

    async def test_sparsewalk_should_have_expected_response(self, snmp_client):
        variables = ("ifIndex", "ifDescr", "ifAlias")
        response = await snmp_client.sparsewalk(*(("IF-MIB", var) for var in variables))
        assert response
        for index, row in response.items():
            assert isinstance(index, OID)
            assert isinstance(row, dict)
            for var, val in row.items():
                assert var in variables

    async def test_get_sysobjectid_should_be_tuple_of_ints(self, snmp_client):
        response = await snmp_client.get("SNMPv2-MIB", "sysObjectID", 0)
        assert isinstance(response.oid, OID)
        assert isinstance(response.value, OID)
        assert all(isinstance(i, int) for i in response.value)

    async def test_get_named_value_should_return_symbolic_name(self, snmp_client):
        response = await snmp_client.getnext("SNMPv2-MIB", "snmpEnableAuthenTraps")
        assert response.value == "disabled"


class TestUnknownMibShouldRaiseException:
    async def test_get(self, snmp_client):
        with pytest.raises(MibNotFoundError):
            await snmp_client.get("NON-EXISTENT-MIB", "foo", 0)

    async def test_getnext(self, snmp_client):
        with pytest.raises(MibNotFoundError):
            await snmp_client.getnext("NON-EXISTENT-MIB", "foo")

    async def test_getnext2(self, snmp_client):
        with pytest.raises(MibNotFoundError):
            await snmp_client.getnext2(("NON-EXISTENT-MIB", "foo"))

    async def test_walk(self, snmp_client):
        with pytest.raises(MibNotFoundError):
            await snmp_client.walk("NON-EXISTENT-MIB", "foo")

    async def test_getbulk(self, snmp_client):
        with pytest.raises(MibNotFoundError):
            await snmp_client.getbulk("NON-EXISTENT-MIB", "foo")

    async def test_bulkwalk(self, snmp_client):
        with pytest.raises(MibNotFoundError):
            await snmp_client.bulkwalk("NON-EXISTENT-MIB", "foo")

    async def test_getbulk2(self, snmp_client):
        with pytest.raises(MibNotFoundError):
            await snmp_client.getbulk2(("NON-EXISTENT-MIB", "foo"))

    async def test_sparsewalk(self, snmp_client):
        with pytest.raises(MibNotFoundError):
            await snmp_client.sparsewalk(("NON-EXISTENT-MIB", "foo"))


class TestMibResolver:
    """Tests to ensure that various required MIBs can be resolved"""

    def test_sysuptime_should_be_resolved(self):
        assert resolve_symbol(("SNMPv2-MIB", "sysUpTime"))

    def test_ifalias_should_be_resolved(self):
        assert resolve_symbol(("IF-MIB", "ifAlias", "1"))

    def test_ipadentaddr_should_be_resolved(self):
        assert resolve_symbol(("IP-MIB", "ipAdEntAddr"))

    def test_jnx_bgp_m2_peer_state_should_be_resolved(self):
        assert resolve_symbol(("BGP4-V2-MIB-JUNIPER", "jnxBgpM2PeerState"))

    def test_c_bgp_peer2_state_should_be_resolved(self):
        assert resolve_symbol(("CISCO-BGP4-MIB", "cbgpPeer2State"))

    def test_bgp_peer_state_should_be_resolved(self):
        assert resolve_symbol(("BGP4-MIB", "bgpPeerState"))

    def test_juniper_bgp_mib_oids_should_be_resolved_to_symbols(self):
        """Ensure vendored MIBs are loaded properly by testing low-level netsnmpy lookup"""
        oid = OID(
            ".1.3.6.1.4.1.2636.5.1.1.2.1.1.1.11.0.2.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.2.32.1.7.0.0.0.128.1.0.0.0.0.0.0.0.2"
        )
        symbol = netsnmp.oid_to_symbol(oid)
        assert symbol.startswith("BGP4-V2-MIB-JUNIPER::jnxBgpM2PeerRemoteAddr")


class TestUnreachableDeviceShouldRaiseException:
    async def test_get(self, unreachable_snmp_client):
        with pytest.raises(TimeoutError):
            await unreachable_snmp_client.get("SNMPv2-MIB", "sysUpTime", 0)

    async def test_getnext(self, unreachable_snmp_client):
        with pytest.raises(TimeoutError):
            await unreachable_snmp_client.getnext("SNMPv2-MIB", "sysUpTime")

    async def test_getnext2(self, unreachable_snmp_client):
        with pytest.raises(TimeoutError):
            await unreachable_snmp_client.getnext2(("SNMPv2-MIB", "sysUpTime"))

    async def test_walk(self, unreachable_snmp_client):
        with pytest.raises(TimeoutError):
            await unreachable_snmp_client.walk("SNMPv2-MIB", "sysUpTime")

    async def test_getbulk(self, unreachable_snmp_client):
        with pytest.raises(TimeoutError):
            await unreachable_snmp_client.getbulk("SNMPv2-MIB", "sysUpTime")

    async def test_bulkwalk(self, unreachable_snmp_client):
        with pytest.raises(TimeoutError):
            await unreachable_snmp_client.bulkwalk("SNMPv2-MIB", "sysUpTime")

    async def test_getbulk2(self, unreachable_snmp_client):
        with pytest.raises(TimeoutError):
            await unreachable_snmp_client.getbulk2(("SNMPv2-MIB", "sysUpTime"))

    async def test_sparsewalk(self, unreachable_snmp_client):
        with pytest.raises(TimeoutError):
            await unreachable_snmp_client.sparsewalk(("SNMPv2-MIB", "sysUpTime"))


async def test_get_object_that_does_not_exist_should_raise_exception(snmp_client):
    with pytest.raises((NoSuchNameError, NoSuchInstanceError)):
        # NoSuchNameError is only relevant on SNMP v1, for v2c the error is NoSuchInstanceError
        await snmp_client.get("SNMPv2-MIB", "sysUpTime", 1)


class TestVarBindErrors:
    """
    Test class for verifying the handling of varbind errors in SNMP commands.

    This class contains tests that check if the correct exceptions are raised when
    varbind errors (NoSuchObject, NoSuchInstance, EndOfMibView) are encountered
    in the response to an SNMP command.
    """

    @pytest.mark.parametrize(
        "error, exception",
        [
            (NoSuchObject(""), NoSuchObjectError),
            (NoSuchInstance(""), NoSuchInstanceError),
            (EndOfMibView(""), EndOfMibViewError),
        ],
        ids=["NoSuchObject-NoSuchObjectError", "NoSuchInstance-NoSuchInstanceError", "EndOfMibView-EndOfMibViewError"],
    )
    async def test_get_should_raise_exception(self, error, exception, snmp_client, monkeypatch):
        query = ("SNMPv2-MIB", "sysDescr", 0)
        oid = resolve_symbol(query)
        mock_results = [SNMPVariable(oid, error)]
        future = asyncio.Future()
        future.set_result(mock_results)
        get_mock = Mock(return_value=future)
        monkeypatch.setattr("zino.snmp.netsnmpy_backend.SNMPSession.aget", get_mock)

        with pytest.raises(exception):
            await snmp_client.get(*query)


class TestSnmpContextManager:
    @pytest.fixture
    def snmp_session(self):
        device = PollDevice(name="localhost", address="127.0.0.1", port=161)
        yield SNMP(device)

    def test_enter_and_exit_should_work(self, snmp_session):
        with snmp_session as snmp:
            assert snmp is snmp_session

    def test_reentrant_contexts_should_not_close_prematurely(self, snmp_session):
        with snmp_session as snmp1:
            assert snmp1._is_open, "first context did not open session"
            with snmp_session as snmp2:
                assert snmp2._is_open, "second context closed the session"
            assert snmp1._is_open, "second session exit close the session prematurely"



#
# PySNMP MIB module JUNIPER-ALARM-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file://./JUNIPER-ALARM-MIB.mib
# Produced by pysmi-1.1.10 at Thu Jul  6 12:36:59 2023
# On host simtve-ThinkPad-X1-Carbon-Gen-9 platform Linux version 5.14.0-1058-oem by user simon
# Using Python version 3.9.16 (main, Dec  7 2022, 01:11:51) 
#
OctetString, ObjectIdentifier, Integer = mibBuilder.importSymbols("ASN1", "OctetString", "ObjectIdentifier", "Integer")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ValueSizeConstraint, ConstraintsIntersection, ValueRangeConstraint, SingleValueConstraint, ConstraintsUnion = mibBuilder.importSymbols("ASN1-REFINEMENT", "ValueSizeConstraint", "ConstraintsIntersection", "ValueRangeConstraint", "SingleValueConstraint", "ConstraintsUnion")
jnxMibs, = mibBuilder.importSymbols("JUNIPER-SMI", "jnxMibs")
ModuleCompliance, NotificationGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ModuleCompliance", "NotificationGroup")
Integer32, Unsigned32, Gauge32, ModuleIdentity, Counter64, ObjectIdentity, iso, IpAddress, MibIdentifier, Bits, MibScalar, MibTable, MibTableRow, MibTableColumn, Counter32, TimeTicks, NotificationType = mibBuilder.importSymbols("SNMPv2-SMI", "Integer32", "Unsigned32", "Gauge32", "ModuleIdentity", "Counter64", "ObjectIdentity", "iso", "IpAddress", "MibIdentifier", "Bits", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "Counter32", "TimeTicks", "NotificationType")
DisplayString, TextualConvention, TimeStamp = mibBuilder.importSymbols("SNMPv2-TC", "DisplayString", "TextualConvention", "TimeStamp")
jnxAlarms = ModuleIdentity((1, 3, 6, 1, 4, 1, 2636, 3, 4))
if mibBuilder.loadTexts: jnxAlarms.setLastUpdated('200307182153Z')
if mibBuilder.loadTexts: jnxAlarms.setOrganization('Juniper Networks, Inc.')
jnxCraftAlarms = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2))
jnxAlarmRelayMode = MibScalar((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2, 1), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3))).clone(namedValues=NamedValues(("other", 1), ("passOn", 2), ("cutOff", 3)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxAlarmRelayMode.setStatus('current')
jnxYellowAlarms = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2, 2))
jnxYellowAlarmState = MibScalar((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2, 2, 1), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3))).clone(namedValues=NamedValues(("other", 1), ("off", 2), ("on", 3)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxYellowAlarmState.setStatus('current')
jnxYellowAlarmCount = MibScalar((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2, 2, 2), Gauge32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxYellowAlarmCount.setStatus('current')
jnxYellowAlarmLastChange = MibScalar((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2, 2, 3), TimeStamp()).setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxYellowAlarmLastChange.setStatus('current')
jnxRedAlarms = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2, 3))
jnxRedAlarmState = MibScalar((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2, 3, 1), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3))).clone(namedValues=NamedValues(("other", 1), ("off", 2), ("on", 3)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxRedAlarmState.setStatus('current')
jnxRedAlarmCount = MibScalar((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2, 3, 2), Gauge32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxRedAlarmCount.setStatus('current')
jnxRedAlarmLastChange = MibScalar((1, 3, 6, 1, 4, 1, 2636, 3, 4, 2, 3, 3), TimeStamp()).setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxRedAlarmLastChange.setStatus('current')
mibBuilder.exportSymbols("JUNIPER-ALARM-MIB", jnxYellowAlarmCount=jnxYellowAlarmCount, jnxRedAlarmLastChange=jnxRedAlarmLastChange, PYSNMP_MODULE_ID=jnxAlarms, jnxAlarms=jnxAlarms, jnxRedAlarmCount=jnxRedAlarmCount, jnxYellowAlarmState=jnxYellowAlarmState, jnxCraftAlarms=jnxCraftAlarms, jnxRedAlarms=jnxRedAlarms, jnxYellowAlarms=jnxYellowAlarms, jnxAlarmRelayMode=jnxAlarmRelayMode, jnxRedAlarmState=jnxRedAlarmState, jnxYellowAlarmLastChange=jnxYellowAlarmLastChange)



import logging
import os
import stat
from unittest.mock import AsyncMock, MagicMock

import aiodns
import pytest

from zino.utils import file_is_world_readable, log_time_spent, reverse_dns


class TestReverseDNS:
    async def test_should_return_reverse_dns_for_valid_ip(self, mock_dnsresolver):
        valid_ip = "8.8.8.8"
        reverse_dns_value = "reverse.dns.example.com"

        # Create mock for return value of gethostbyaddr
        mock_gethostbyaddr_return_value = MagicMock()
        mock_gethostbyaddr_return_value.configure_mock(name=reverse_dns_value, aliases=[], addresses=[valid_ip])

        # Create gethostbyaddr async mock
        mock_gethostbyaddr = AsyncMock(return_value=mock_gethostbyaddr_return_value)

        # Mock DNSRsolver.gethostbyaddr function
        mock_dnsresolver.gethostbyaddr = mock_gethostbyaddr

        result = await reverse_dns(valid_ip)
        assert result == reverse_dns_value

    async def test_should_return_none_for_invalid_ip(self, mock_dnsresolver):
        invalid_ip = "0.0.0.0"

        # Mock DNSResolver.gethostbyaddr to raise DNSError
        mock_gethostbyaddr = AsyncMock(side_effect=aiodns.error.DNSError)
        mock_dnsresolver.gethostbyaddr = mock_gethostbyaddr

        result = await reverse_dns(invalid_ip)
        assert result is None


class TestLogTimeSpent:
    def test_when_logger_is_specified_it_should_log_time_spent_using_that_logger(self, caplog):
        @log_time_spent(logger="test_logger", level=logging.DEBUG)
        def test_function():
            pass

        with caplog.at_level(logging.DEBUG):
            test_function()

        assert any(
            record.name == "test_logger"
            and "took" in record.msg
            and "seconds" in record.msg
            and "test_function" in record.args
            for record in caplog.records
        )


@pytest.fixture
def mock_dnsresolver(monkeypatch) -> AsyncMock:
    mock_dnsresolver = AsyncMock()
    monkeypatch.setattr("zino.utils.aiodns.DNSResolver", lambda loop: mock_dnsresolver)
    return mock_dnsresolver


class TestFileIsReadableByOthers:
    def test_return_true_if_file_is_world_readable(self, secrets_file):
        assert file_is_world_readable(secrets_file)

    def test_return_if_file_is_only_readable_by_owner(self, tmp_path):
        name = tmp_path / "owner-secrets"
        with open(name, "w") as conf:
            conf.write("""user1 password123""")
        os.chmod(name, mode=stat.S_IRWXU)

        assert not file_is_world_readable(name)



#
# PySNMP MIB module JUNIPER-BFD-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file://./JUNIPER-BFD-MIB.mib
# Produced by pysmi-1.1.10 at Mon Oct  9 12:14:08 2023
# On host simtve-ThinkPad-X1-Carbon-Gen-9 platform Linux version 5.15.0-84-generic by user simon
# Using Python version 3.9.18 (main, Aug 25 2023, 13:20:04) 
#
Integer, ObjectIdentifier, OctetString = mibBuilder.importSymbols("ASN1", "Integer", "ObjectIdentifier", "OctetString")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsUnion, ValueRangeConstraint, ValueSizeConstraint, ConstraintsIntersection, SingleValueConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsUnion", "ValueRangeConstraint", "ValueSizeConstraint", "ConstraintsIntersection", "SingleValueConstraint")
bfdSessIndex, = mibBuilder.importSymbols("BFD-STD-MIB", "bfdSessIndex")
jnxBfdMibRoot, = mibBuilder.importSymbols("JUNIPER-SMI", "jnxBfdMibRoot")
NotificationGroup, ModuleCompliance = mibBuilder.importSymbols("SNMPv2-CONF", "NotificationGroup", "ModuleCompliance")
IpAddress, ModuleIdentity, Bits, MibIdentifier, Counter64, iso, Integer32, TimeTicks, NotificationType, Unsigned32, MibScalar, MibTable, MibTableRow, MibTableColumn, ObjectIdentity, Gauge32, Counter32 = mibBuilder.importSymbols("SNMPv2-SMI", "IpAddress", "ModuleIdentity", "Bits", "MibIdentifier", "Counter64", "iso", "Integer32", "TimeTicks", "NotificationType", "Unsigned32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "ObjectIdentity", "Gauge32", "Counter32")
TimeStamp, DisplayString, TruthValue, TextualConvention = mibBuilder.importSymbols("SNMPv2-TC", "TimeStamp", "DisplayString", "TruthValue", "TextualConvention")
jnxBfdMib = ModuleIdentity((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1))
jnxBfdMib.setRevisions(('2006-10-12 12:00',))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    if mibBuilder.loadTexts: jnxBfdMib.setRevisionsDescriptions(('Initial version.',))
if mibBuilder.loadTexts: jnxBfdMib.setLastUpdated('200610121200Z')
if mibBuilder.loadTexts: jnxBfdMib.setOrganization('IETF')
if mibBuilder.loadTexts: jnxBfdMib.setContactInfo(' Juniper Technical Assistance Center Juniper Networks, Inc. 1194 N. Mathilda Avenue Sunnyvale, CA 94089 E-mail: support@juniper.net')
if mibBuilder.loadTexts: jnxBfdMib.setDescription('Provides BFD specific data.')
jnxBfdNotification = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 0))
jnxBfdObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 1))
jnxBfdNotifyObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 2))
jnxBfdSessTable = MibTable((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 1, 1), )
if mibBuilder.loadTexts: jnxBfdSessTable.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessTable.setDescription('Defines the jnxBfd Session Table for providing enterprise specific options to the corresponding bfdSessTable entry.')
jnxBfdSessEntry = MibTableRow((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 1, 1, 1), ).setIndexNames((0, "BFD-STD-MIB", "bfdSessIndex"))
if mibBuilder.loadTexts: jnxBfdSessEntry.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessEntry.setDescription('Defines an entry in the jnxBfdSessTable. This essentially augments the bfdSessTable with additional objects.')
jnxBfdSessThreshTxInterval = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 1, 1, 1, 1), Unsigned32()).setUnits('microseconds').setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxBfdSessThreshTxInterval.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessThreshTxInterval.setDescription('The threshold value for transmit interval in microseconds. If the current transmit interval value adapts to a value greater than the threshold value, jnxBfdSessTxIntervalHigh trap is raised.')
jnxBfdSessCurrTxInterval = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 1, 1, 1, 2), Unsigned32()).setUnits('microseconds').setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxBfdSessCurrTxInterval.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessCurrTxInterval.setDescription('The current transmit interval in microseconds.')
jnxBfdSessThreshDectTime = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 1, 1, 1, 3), Unsigned32()).setUnits('microseconds').setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxBfdSessThreshDectTime.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessThreshDectTime.setDescription('The threshold value for detection time in microseconds. If the current detection time value is greater than the threshold value at the time when session state changes to up(1), jnxBfdSessDetectionTimeHigh trap is raised.')
jnxBfdSessCurrDectTime = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 1, 1, 1, 4), Unsigned32()).setUnits('microseconds').setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxBfdSessCurrDectTime.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessCurrDectTime.setDescription('The actual value of detection time for the session.')
jnxBfdSessIntfName = MibTableColumn((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 1, 1, 1, 5), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: jnxBfdSessIntfName.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessIntfName.setDescription('This object specifies the interface associated with the bfd session')
jnxBfdSessifName = MibScalar((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 2, 1), DisplayString()).setMaxAccess("accessiblefornotify")
if mibBuilder.loadTexts: jnxBfdSessifName.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessifName.setDescription('This object gives the Interface Name in the bfdSessUp and bfdSessDown trap. Even though this object doesnt appear in the OBJECTS list of these traps, but the agent relay this information as an extra parameter in the trap.')
jnxBfdSessTxIntervalHigh = NotificationType((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 0, 1)).setObjects(("JUNIPER-BFD-MIB", "jnxBfdSessThreshTxInterval"), ("JUNIPER-BFD-MIB", "jnxBfdSessCurrTxInterval"))
if mibBuilder.loadTexts: jnxBfdSessTxIntervalHigh.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessTxIntervalHigh.setDescription('This notification is generated when the threshold value for transmit interval (jnxBfdSessThreshTxInterval) is set and the bfd session transmit interval (jnxBfdSessCurrTxInterval) adapts to a value greater than the threshold value. This trap is sent only once, when we first exceed the threshold. The transmit interval can continue to adapt beyond the threshold value. Adaptation of transmit interval happens due to network issues causing the BFD session to go down on either the local system or the remote neighbor.')
jnxBfdSessDetectionTimeHigh = NotificationType((1, 3, 6, 1, 4, 1, 2636, 3, 45, 1, 0, 2)).setObjects(("JUNIPER-BFD-MIB", "jnxBfdSessThreshDectTime"), ("JUNIPER-BFD-MIB", "jnxBfdSessCurrDectTime"))
if mibBuilder.loadTexts: jnxBfdSessDetectionTimeHigh.setStatus('current')
if mibBuilder.loadTexts: jnxBfdSessDetectionTimeHigh.setDescription('This notification is generated when the threshold value for detection time (jnxBfdSessThreshDectTime) is set and the bfd session detection-time (jnxBfdSessCurrDectTime) adapts to a value greater than the threshold value. This trap is sent only once, when we first exceed the threshold. The detection-time can continue to adapt beyond the threshold value. Adaptation of detection-time happens due to network issues causing the BFD session to go down on either the local system or the remote neighbor.')
mibBuilder.exportSymbols("JUNIPER-BFD-MIB", jnxBfdSessIntfName=jnxBfdSessIntfName, jnxBfdSessThreshTxInterval=jnxBfdSessThreshTxInterval, jnxBfdSessTable=jnxBfdSessTable, jnxBfdMib=jnxBfdMib, jnxBfdSessThreshDectTime=jnxBfdSessThreshDectTime, jnxBfdObjects=jnxBfdObjects, jnxBfdSessifName=jnxBfdSessifName, jnxBfdSessTxIntervalHigh=jnxBfdSessTxIntervalHigh, jnxBfdNotification=jnxBfdNotification, jnxBfdSessDetectionTimeHigh=jnxBfdSessDetectionTimeHigh, jnxBfdSessEntry=jnxBfdSessEntry, PYSNMP_MODULE_ID=jnxBfdMib, jnxBfdSessCurrDectTime=jnxBfdSessCurrDectTime, jnxBfdSessCurrTxInterval=jnxBfdSessCurrTxInterval, jnxBfdNotifyObjects=jnxBfdNotifyObjects)



"""This module implements custom logging of certain SNMP traps, but takes no other action"""

import logging
from ipaddress import ip_address
from typing import Optional

from zino.trapd.base import TrapMessage, TrapObserver

_logger = logging.getLogger(__name__)


class RestartTrapLogger(TrapObserver):
    WANTED_TRAPS = {
        ("SNMPv2-MIB", "coldStart"),
        ("SNMPv2-MIB", "warmStart"),
    }

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        _logger.info("%s: %s", trap.agent.device.name, trap.name)
        return False  # stop trap processing here


class CiscoReloadTrapLogger(TrapObserver):
    WANTED_TRAPS = {
        ("CISCOTRAP-MIB", "reload"),
    }

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        _logger.info("%s: reload requested", trap.agent.device.name)
        return False  # stop trap processing here


class CiscoConfigManEventLogger(TrapObserver):
    WANTED_TRAPS = {
        ("CISCO-CONFIG-MAN", "ciscoConfigManEvent"),
    }

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        _logger.info(
            "%s: config-change: cmd-src %s conf-src %s dst %s",
            trap.agent.device.name,
            *[
                trap.get_all(var)[0].value if var in trap else None
                for var in (
                    "ccmHistoryEventCommandSource",
                    "ccmHistoryEventConfigSource",
                    "ccmHistoryEventConfigDestination",
                )
            ],
        )

        return False  # stop trap processing here


class CiscoPimTrapLogger(TrapObserver):
    """Logs details of Cisco PIM traps.

    Example of manually generating a matching trap with test data:

          snmptrap -v 2c -c public localhost:1162 "" \
               CISCO-PIM-MIB::ciscoPimInvalidRegister \
               CISCO-PIM-MIB::cpimLastErrorOriginType i 1 \
               CISCO-PIM-MIB::cpimLastErrorOrigin x "0A000001" \
               CISCO-PIM-MIB::cpimLastErrorGroupType i 1 \
               CISCO-PIM-MIB::cpimLastErrorGroup x "0A000002" \
               CISCO-PIM-MIB::cpimLastErrorRPType i 1 \
               CISCO-PIM-MIB::cpimLastErrorRP  x "0A000003" \
               CISCO-PIM-MIB::cpimInvalidRegisterMsgsRcvd c 42
    """

    WANTED_TRAPS = {
        ("CISCO-PIM-MIB", "ciscoPimInvalidRegister"),
        ("CISCO-PIM-MIB", "ciscoPimInvalidJoinPrune"),
    }

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        if "cpimLastErrorOriginType" not in trap:
            return False

        if trap.get_all("cpimLastErrorOriginType")[0].value != "ipv4":
            return False

        origin = ip_address(bytes(trap.get_all("cpimLastErrorOrigin")[0].raw_value))
        group = ip_address(bytes(trap.get_all("cpimLastErrorGroup")[0].raw_value))
        rp = ip_address(bytes(trap.get_all("cpimLastErrorRP")[0].raw_value))

        trap_type = "register" if trap.name.endswith("Register") else "join-prune"
        _logger.info(
            "%s PIM-invalid-%s: from %s group %s RP %s",
            trap.agent.device.name,
            trap_type,
            origin,
            group,
            rp,
        )

        return False  # stop trap processing here


class OspfIfConfigErrorLogger(TrapObserver):
    """Logs all ospfIfConfigError trap variables prefixed by 'ospf*'"""

    WANTED_TRAPS = {
        ("OSPF-MIB", "ospfIfConfigError"),
    }

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        _logger.info("%s: trap %s", trap.agent.device.name, trap.get_all("snmpTrapOID")[0].value)

        for var in trap.variables:
            if var.var.startswith("ospf"):
                _logger.info("%s: trap-var %s: %s", trap.agent.device.name, var.var, var.value)

        return False  # stop trap processing here



import os
from datetime import datetime, timedelta, timezone
from ipaddress import ip_address

import pytest

from zino.stateconverter.convertstate import create_state
from zino.statemodels import (
    AlarmEvent,
    BFDEvent,
    BFDSessState,
    BGPAdminStatus,
    BGPEvent,
    BGPOperState,
    EventState,
    InterfaceState,
    MatchType,
    PmType,
    PortStateEvent,
    PortStateMaintenance,
)


def test_state_should_be_parsed_without_crashing(save_state_path):
    state = create_state(save_state_path)
    assert state


class TestEvents:
    def test_bfd_event_should_be_created_correctly(self, save_state_path):
        state = create_state(save_state_path)
        event = state.events.checkout(200)
        assert isinstance(event, BFDEvent)
        assert event.type == "bfd"
        assert event.bfdaddr == ip_address("219.188.192.78")
        assert event.bfddiscr == 4500
        assert event.bfdix == 30
        assert event.bfdstate == BFDSessState.DOWN
        assert len(event.history) == 1
        assert len(event.log) == 1
        assert event.lastevent == "changed from up to adminDown (poll)"
        assert event.opened == datetime.fromtimestamp(1700400123, tz=timezone.utc)
        assert event.polladdr == ip_address("93.150.77.115")
        assert event.priority == 100
        assert event.router == "blaafjell-gw2"
        assert event.state == EventState.OPEN
        assert event.updated == datetime.fromtimestamp(1700400123, tz=timezone.utc)
        assert event.neigh_rdns == "nissen.nordpolen.no"

    def test_alarm_event_should_be_created_correctly(self, save_state_path):
        state = create_state(save_state_path)
        event = state.events.checkout(146)
        assert isinstance(event, AlarmEvent)
        assert event.type == "alarm"
        assert event.alarm_type == "yellow"
        assert event.alarm_count == 1
        assert len(event.history) == 2
        assert len(event.log) == 2
        assert event.lastevent == "alarms went from 0 to 1"
        assert event.opened == datetime.fromtimestamp(1696257668, tz=timezone.utc)
        assert event.polladdr == ip_address("176.53.125.80")
        assert event.priority == 300
        assert event.router == "whoville-gw1"
        assert event.state == EventState.WAITING
        assert event.updated == datetime.fromtimestamp(1696257668, tz=timezone.utc)

    def test_portstate_event_should_be_created_correctly(self, save_state_path):
        state = create_state(save_state_path)
        event = state.events.checkout(110)
        assert isinstance(event, PortStateEvent)
        assert event.type == "portstate"
        assert event.ifindex == 150
        assert event.descr == "LACP-link, test.no-phy1"
        assert event.ac_down == timedelta(seconds=15000000)
        assert len(event.history) == 1
        assert len(event.log) == 1
        assert event.opened == datetime.fromtimestamp(1686257668, tz=timezone.utc)
        assert event.polladdr == ip_address("53.44.228.67")
        assert event.port == "ge-1/0/10"
        assert event.portstate == InterfaceState.UP
        assert event.priority == 100
        assert event.router == "arkham-sw1"
        assert event.state == EventState.IGNORED
        assert event.updated == datetime.fromtimestamp(1686257668, tz=timezone.utc)

    def test_bgp_event_should_be_created_correctly(self, save_state_path):
        """Placeholder for when BGP is supported"""
        state = create_state(save_state_path)
        event = state.events.checkout(100)
        assert isinstance(event, BGPEvent)
        assert event.type == "bgp"
        assert event.router == "auroralane-gw1"
        assert event.state == EventState.WAITING
        assert event.polladdr == ip_address("219.188.192.78")
        assert event.opened == datetime.fromtimestamp(1706257668, tz=timezone.utc)
        assert event.updated == datetime.fromtimestamp(1706257668, tz=timezone.utc)
        assert event.priority == 100
        assert len(event.history) == 1
        assert len(event.log) == 1
        assert event.lastevent == "peer is down"
        assert event.remote_address == ip_address("0515:7cfd:1bcc:279e:f5a4:5528:f4c3:58af")
        assert event.remote_as == 100
        assert event.peer_uptime == 420
        assert event.operational_state == BGPOperState.DOWN
        assert event.admin_status == BGPAdminStatus.RUNNING

    def test_invalid_event_attribute_should_not_be_set(self, invalid_event_save_state_path):
        state = create_state(invalid_event_save_state_path)
        event = state.events.checkout(100)
        assert not hasattr(event, "invalid_attr")


def test_jnx_alarms_should_be_set_correctly(save_state_path):
    state = create_state(save_state_path)
    device = state.devices.get("whoville-gw1")
    assert device.alarms["yellow"] == 1
    assert device.alarms["red"] == 0


class TestBFD:
    def test_sess_addr_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("blaafjell-gw2")
        device.ports[30].bfd_state.session_addr is None

    def test_sess_discr_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("blaafjell-gw2")
        device.ports[30].bfd_state.session_discr == 0

    def test_sess_state_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("blaafjell-gw2")
        device.ports[30].bfd_state.session_state == BFDSessState.DOWN


class TestVendor:
    def test_juniper_device_should_be_registered_as_juniper(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("juniper-gw1")
        assert device.is_juniper

    def test_cisco_device_should_be_registered_as_cisco(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("cisco-gw1")
        assert device.is_cisco


class TestPort:
    def test_portstate_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("arkham-sw1")
        assert device.ports[150].state == InterfaceState.UP

    def test_ifdescr_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("arkham-sw1")
        assert device.ports[150].ifdescr == "ge-1/0/10"

    def test_ifalias_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("arkham-sw1")
        assert device.ports[150].ifalias == "LACP-link, test.no-phy1"


class TestBGP:
    def test_uptime_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("auroralane-gw1")
        ip = ip_address("3000:04AB:0554:0001:0000:0000:0000:00AA")
        assert device.bgp_peers[ip].uptime == 14000000

    def test_admin_status_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("auroralane-gw1")
        ip = ip_address("3000:04AB:0554:0001:0000:0000:0000:00AA")
        assert device.bgp_peers[ip].admin_status == BGPAdminStatus.RUNNING

    def test_oper_status_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("auroralane-gw1")
        ip = ip_address("3000:04AB:0554:0001:0000:0000:0000:00AA")
        assert device.bgp_peers[ip].oper_state == BGPOperState.ACTIVE

    def test_peers_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        device = state.devices.get("auroralane-gw1")
        ip = ip_address("3000:04AB:0554:0001:0000:0000:0000:00AA")
        assert ip in device.bgp_peers
        assert len(device.bgp_peers) == 1


class TestPM:
    def test_pm_should_be_created_correctly(self, save_state_path):
        state = create_state(save_state_path)
        pm = state.planned_maintenances[3188]
        assert isinstance(pm, PortStateMaintenance)
        assert pm.type == PmType.PORTSTATE
        assert pm.start_time == datetime.fromtimestamp(1720021526, tz=timezone.utc)
        assert pm.end_time == datetime.fromtimestamp(1720025126, tz=timezone.utc)
        assert pm.match_type == MatchType.INTF_REGEXP
        assert pm.match_expression == "ge-1/0/10"
        assert pm.match_device == "blaafjell-gw2"
        assert pm.event_ids == [110]
        assert len(pm.log) == 1

    def test_last_pm_id_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        assert state.planned_maintenances.last_pm_id == 3188

    def test_last_run_should_be_set_correctly(self, save_state_path):
        state = create_state(save_state_path)
        assert state.planned_maintenances.last_run == datetime.fromtimestamp(1720018082, tz=timezone.utc)


def test_addresses_should_be_set_correctly(save_state_path):
    state = create_state(save_state_path)
    ip = ip_address("175.46.88.27")
    assert state.addresses[ip] == "boot-gw1"


@pytest.fixture
def save_state_path():
    this_directory = os.path.dirname(__file__)
    state_file = os.path.join(this_directory, "fixtures", "zino1_state.tcl")
    return state_file


@pytest.fixture
def invalid_event_save_state_path():
    this_directory = os.path.dirname(__file__)
    state_file = os.path.join(this_directory, "fixtures", "zino1_state_invalid_event_attr.tcl")
    return state_file



from dataclasses import dataclass
from typing import Optional


@dataclass
class LineData:
    """Flexible definition of the value and identifiers found in a Zino1 .tcl state dump.
    Each line in the state dump can contain several identifiers but there is always only one value.
    """

    identifiers: Optional[tuple[str, ...]]
    value: str


def get_line_data(line) -> LineData:
    """Parses a line from a Zino1 .tcl filedump into a LineData object
    containing useful information
    """
    try:
        identifiers = get_identifiers(line)
    except IndexError:
        identifiers = None
    value = get_value(line)
    return LineData(value=value, identifiers=identifiers)


def get_identifiers(line: str) -> tuple[str, ...]:
    # removes part of line before identifiers are defined
    split_line = line.split("(")[1]
    # removes everything after the identifiers
    split_line = split_line.split(")")[0]
    identifiers = split_line.split(",")
    if line.startswith("set ::EventAttrs_") or line.startswith("set ::pm::event_"):
        # remove everything before the event ID starts
        event_line = line.split("_")[1]
        # remove everything after event ID
        event_id = event_line.split("(")[0]
        identifiers.append(event_id)
    return tuple(identifiers)


def get_value(line: str) -> str:
    # remove everything before the value is defined
    value = line.split(" ")[2:]
    value = " ".join(value)
    # strip whitespace and quotes
    value = value.strip(' "')
    return value



"""Tests for functions that are unchanged in subclasses of PlannedMaintenance"""

import datetime

import pytest

from zino.statemodels import (
    EventState,
    MatchType,
    PlannedMaintenance,
    PmType,
    ReachabilityEvent,
)


class TestEnd:
    def test_should_set_state_to_open_for_all_events_in_events_id_list(self, state, reachability_event, pm):
        reachability_event.state = EventState.IGNORED
        state.events.commit(reachability_event)
        pm.event_ids.append(reachability_event.id)
        pm.end(state)
        assert state.events.get(reachability_event.router, None, ReachabilityEvent).state is EventState.OPEN


class TestAddLog:
    def test_should_add_log_entry_to_log_list_with_correct_message(self, pm):
        msg = "log msg"
        entry = pm.add_log(msg)
        assert entry.message == msg
        assert entry in pm.log


@pytest.fixture
def pm() -> PlannedMaintenance:
    return PlannedMaintenance(
        start_time=datetime.datetime.now() - datetime.timedelta(days=1),
        end_time=datetime.datetime.now() + datetime.timedelta(days=1),
        type=PmType.DEVICE,
        match_type=MatchType.STR,
        match_expression="device",
        match_device=None,
    )


@pytest.fixture
def regexp_pm() -> PlannedMaintenance:
    return PlannedMaintenance(
        start_time=datetime.datetime.now() - datetime.timedelta(days=1),
        end_time=datetime.datetime.now() + datetime.timedelta(days=1),
        type=PmType.DEVICE,
        match_type=MatchType.REGEXP,
        match_expression="device",
        match_device=None,
    )
